# -*- coding: utf-8 -*-

import os
import re

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np

CURRENT_DIR = os.path.realpath(os.path.basename(__file__))

# <PROFILING>SenderFederate2[131074](initializing)HELICS CODE ENTRY<4570827706580384>[t=-1000000]</PROFILING>
PATTERN = re.compile(
    r"""
            (?P<name>\w+)                           # SenderFederate2
            \[(\d+)\]                               # [131074]
            \((?P<state>\w+)\)                      # (initializing)
            (?P<message>(?:\w|\s)+)                 # HELICS CODE ENTRY
            \<(?P<realtime>\d+(?:\|\d+)?)\>         # <4570827706580384|534534523453>
            \[t=(?P<simtime>-?\d*\.{0,1}\d+)\]      # [t=-1000000]
            """,
    re.X,
)


def profile(filename, invert=True):
    with open(filename) as f:
        data = f.read()
    data = data.replace(r"<PROFILING>", "").replace(r"</PROFILING>", "")
    names = []
    states = []
    messages = []
    simtimes = []
    realtimes = []
    time_marker = {}
    for line in data.splitlines():
        m = PATTERN.match(line)
        name = m.group("name")
        state = m.group("state")
        message = m.group("message")
        simtime = float(m.group("simtime"))
        try:
            realtime = float(m.group("realtime"))
        except:
            realtime, markertime = m.group("realtime").split("|")
            time_marker[name] = float(markertime)
            realtime = float(realtime)
        names.append(name)
        states.append(state)
        messages.append(message)
        simtimes.append(simtime)
        realtimes.append(realtime)

    profile = {}
    for name in set(names):
        profile[name] = []

    if invert:
        for name in set(names):
            profile[name].append({})

    for (name, state, message, simtime, realtime) in zip(names, states, messages, simtimes, realtimes):
        if state == "created":
            continue
        if "ENTRY" in message and not invert:
            profile[name].append({"s_enter": simtime, "r_enter": realtime})
        elif "EXIT" in message and not invert:
            profile[name][-1]["s_end"] = simtime
            profile[name][-1]["r_end"] = realtime
        elif "EXIT" in message and invert:
            profile[name].append({"s_enter": simtime, "r_enter": realtime})
        elif "ENTRY" in message and invert:
            profile[name][-1]["s_end"] = simtime
            profile[name][-1]["r_end"] = realtime

    return profile


def plot(profile, save=None, kind="simulation", **kwargs):
    profiles = []
    names = {k: i for i, k in enumerate(sorted(profile.keys()))}

    if kind == "simulation":
        end = "s_end"
        enter = "s_enter"
        scaling = 1
    elif kind == "realtime":
        end = "r_end"
        enter = "r_enter"
        scaling = 1e9
    else:
        raise Exception("unknown kind")

    for k in profile.keys():
        for i in profile[k]:
            if end in i.keys() and enter in i.keys():
                i["name"] = k
                i[enter] = i[enter] / scaling
                i[end] = i[end] / scaling
                profiles.append(i)

    fig, axs = plt.subplots(1, 1, figsize=(16, 9))
    ax = axs

    cmap = matplotlib.cm.get_cmap("seismic")

    values = list(d[end] - d[enter] for d in profiles)
    norm = matplotlib.colors.Normalize(vmin=min(values), vmax=max(values))

    m_enter = min(d[enter] for d in profiles)
    for i, d in enumerate(profiles):
        d["color"] = cmap(norm(values[i]))

    for d in profiles:
        d[enter] = d[enter] - m_enter
        d[end] = d[end] - m_enter
        ax.barh(names[d["name"]], d[end] - d[enter], left=d[enter], edgecolor=d["color"], color=d["color"], **kwargs)

    if kind == "Simulation":
        ax.set_xlabel("Simulation Time (s)")
    else:
        ax.set_xlabel("Real Time (s)")
    ax.set_yticks([i for i in range(0, len(names))])
    ax.set_yticklabels(sorted(names.keys()))
    ax.set_facecolor("#f0f0f0")
    fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax, location="top")
    if save is None:
        plt.show()
    else:
        plt.savefig(save, dpi=300)


if __name__ == "__main__":
    plot(profile(os.path.abspath(os.path.join(CURRENT_DIR, "../examples/pi-exchange/profile.txt"))), kind="realtime")



# -*- coding: utf-8 -*-
import time
import sys

from math import pi
import helics as h
import random

federate_name = f"SenderFederate{sys.argv[1]}"

print(f"{federate_name}: Helics version = {h.helicsGetVersion()}")

fedinfo = h.helicsCreateFederateInfo()

h.helicsFederateInfoSetCoreName(fedinfo, federate_name + "Core")
h.helicsFederateInfoSetCoreTypeFromString(fedinfo, "zmq")
h.helicsFederateInfoSetCoreInitString(fedinfo, "--federates=1")
h.helicsFederateInfoSetTimeProperty(fedinfo, h.helics_property_time_delta, 1.0)
h.helicsFederateInfoSetTimeProperty(fedinfo, h.helics_property_time_period, 1.0)

fed = h.helicsCreateCombinationFederate(federate_name, fedinfo)

print(f"{federate_name}: Combination federate created")

pub = h.helicsFederateRegisterGlobalTypePublication(fed, f"globaltopic{sys.argv[1]}", "double", "")

print(f"{federate_name}: Publication registered")

h.helicsFederateEnterExecutingMode(fed)

print(f"{federate_name}: Entering execution mode")

this_time = 0.0
grantedtime = -1

for t in range(1, 10):
    value = pi * t

    while grantedtime < t:
        grantedtime = h.helicsFederateRequestTime(fed, t)

    h.helicsPublicationPublishDouble(pub, value)

    print(f"{federate_name}: Sending value pi = {value} at time {grantedtime}")

    # Computing user needs
    time.sleep(float(sys.argv[1]) * (1 + (0.5 - random.random()) / 10))

h.helicsFederateFinalize(fed)
print(f"{federate_name}: Federate finalized")

h.helicsFederateFree(fed)
h.helicsCloseLibrary()



# -*- coding: utf-8 -*-
import logging
import json
import time
import os
import shlex
import subprocess
import sys
from dataclasses import dataclass
from typing import cast

from flask import Flask, render_template, send_from_directory, request, jsonify
from flask_restful import Resource, Api, reqparse, abort
from flask_cors import CORS
import sqlalchemy as sa
from sqlalchemy.ext.automap import automap_base
import werkzeug

import re

from .. import database as db

current_directory = os.path.realpath(os.path.dirname(__file__))

app = Flask(__name__.split(".")[0], static_url_path="", static_folder=os.path.join(current_directory, "../static"))
api = Api(app)
CORS(app, resources={r"/api/*": {"origins": "*"}})

app.config["UPLOAD_FOLDER"] = "/tmp/uploads"

cache = {
    "path": os.path.join(app.config["UPLOAD_FOLDER"], "helics-cli.sqlite.db"),
    "profile-path": os.path.join(app.config["UPLOAD_FOLDER"], "profile.txt"),
    "runner-path": os.path.join(app.config["UPLOAD_FOLDER"], "runner.json"),
    "runner-folder": app.config["UPLOAD_FOLDER"],
    "runner-file-name": "runner.json",
}


class DatabaseManager:
    def __init__(self, path_to_helics_db=cache["path"]):
        self.path_to_helics_db = path_to_helics_db
        self.instaniate()

    def instaniate(self):
        self.engine = db.create_engine(f"sqlite:///{self.path_to_helics_db}")
        self.connection = self.engine.connect()
        self.session = db.Session(self.engine)


class Database(Resource):
    def get(self):
        return cache["path"]

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("file", type=werkzeug.datastructures.FileStorage, location="files")
        args = parser.parse_args()
        file = args["file"]
        os.makedirs(app.config["UPLOAD_FOLDER"], exist_ok=True)
        file.save(os.path.join(app.config["UPLOAD_FOLDER"], "helics-cli.sqlite.db"))
        cache["path"] = os.path.join(app.config["UPLOAD_FOLDER"], "helics-cli.sqlite.db")
        dm = DatabaseManager()
        return {"filename": dm.path_to_helics_db}


api.add_resource(Database, "/api/observer/database")


class SystemInfo(Resource):
    def get(self):
        dm = DatabaseManager()
        return dm.session.query(db.SystemInfo).one().data


api.add_resource(SystemInfo, "/api/observer/systeminfo")


class Cores(Resource):
    def get(self):
        dm = DatabaseManager()
        cores = dm.session.query(db.Cores).all()
        return [{"id": e.id, "name": e.name, "address": e.address} for e in cores if not e.name.startswith("__observer__")]


api.add_resource(Cores, "/api/observer/cores")


class Federates(Resource):
    def get(self):
        dm = DatabaseManager()
        federates = dm.session.query(db.Federates).all()
        return [{"id": f.id, "name": f.name, "parent": f.parent} for f in federates if f.name != "__observer__"]


api.add_resource(Federates, "/api/observer/federates")


class Graph(Resource):
    def get(self):
        dm = DatabaseManager()
        federate = dm.session.query(db.FederateGraph).one().data
        data = dm.session.query(db.DataGraph).one().data
        return {"federate": federate, "data": data}


api.add_resource(Graph, "/api/observer/graphs")


class Subscriptions(Resource):
    def get(self):
        dm = DatabaseManager()
        subscriptions = dm.session.query(db.Subscriptions).all()
        return [db.as_dict(i) for i in subscriptions]


api.add_resource(Subscriptions, "/api/observer/subscriptions")


class Inputs(Resource):
    def get(self):
        dm = DatabaseManager()
        inputs = dm.session.query(db.Inputs).all()
        return [db.as_dict(i) for i in inputs]


api.add_resource(Inputs, "/api/observer/inputs")


class Publications(Resource):
    def get(self):
        dm = DatabaseManager()
        publications = dm.session.query(db.Publications).all()
        return [db.as_dict(i) for i in publications]


api.add_resource(Publications, "/api/observer/publications")


class DataTable(Resource):
    def get(self):
        dm = DatabaseManager()
        Base = automap_base()
        Base.prepare(dm.engine, reflect=True)
        return [db.as_dict(i) for i in dm.session.query(Base.classes.datatable).all()]


api.add_resource(DataTable, "/api/observer/data")

status_tracker = {}


class RunnerFile(Resource):
    def get(self):
        with open(cache["runner-path"]) as f:
            data = json.loads(f.read())
        data["folder"] = cache["runner-folder"]
        data["path"] = cache["runner-path"]
        data["filename"] = cache["runner-file-name"]
        for federate in data["federates"]:
            federate["old_name"] = federate["name"]
            if os.path.exists(os.path.join(data["folder"], "{}.log".format(federate["name"]))):
                federate["log_available"] = True
            else:
                federate["log_available"] = False

            if federate["name"] in status_tracker:
                federate["status"] = status_tracker[federate["name"]]
            else:
                federate["status"] = None
        return data

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("file", type=werkzeug.datastructures.FileStorage, location="files")
        args = parser.parse_args()
        file = args["file"]
        path = cache["runner-folder"]
        name = cache["runner-file-name"]
        os.makedirs(path, exist_ok=True)
        file.save(os.path.join(path, name))
        cache["runner-path"] = os.path.join(path, name)


api.add_resource(RunnerFile, "/api/runner/file")


class RunnerFileName(Resource):
    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("name", type=str)
        args = parser.parse_args()
        name = args["name"]
        cache["runner-file-name"] = name
        cache["runner-path"] = os.path.join(cache["runner-folder"], cache["runner-file-name"])


api.add_resource(RunnerFileName, "/api/runner/file/name")


class RunnerFileFolder(Resource):
    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("folder", type=str)
        args = parser.parse_args()
        folder = args["folder"]
        cache["runner-folder"] = os.path.abspath(os.path.expanduser(folder))
        cache["runner-path"] = os.path.join(cache["runner-folder"], cache["runner-file-name"])


api.add_resource(RunnerFileFolder, "/api/runner/file/folder")


class RunnerFilePath(Resource):
    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("path", type=str)
        args = parser.parse_args()
        path = os.path.abspath(os.path.expanduser(args["path"]))
        cache["runner-path"] = path
        cache["runner-folder"] = os.path.dirname(path)
        cache["runner-file-name"] = os.path.basename(path)


api.add_resource(RunnerFilePath, "/api/runner/file/path")


class RunnerFileEdit(Resource):
    def delete(self):
        parser = reqparse.RequestParser()
        parser.add_argument("old_name", type=str)
        args = parser.parse_args()
        old_name = args["old_name"]
        with open(cache["runner-path"]) as f:
            data = json.loads(f.read())

        delete_index = None
        for i, f in enumerate(data["federates"]):
            if f["name"] == old_name:
                delete_index = i
                break
        data["federates"].pop(delete_index)
        with open(cache["runner-path"], "w") as f:
            f.write(json.dumps(data))
        return {"status": 200}

    def put(self):
        parser = reqparse.RequestParser()
        parser.add_argument("old_name", type=str)
        parser.add_argument("name", type=str)
        parser.add_argument("exec", type=str)
        parser.add_argument("directory", type=str)
        args = parser.parse_args()
        old_name = args["old_name"]
        name = args["name"]
        exec = args["exec"]
        directory = args["directory"]

        with open(cache["runner-path"]) as f:
            data = json.loads(f.read())

        for f in data["federates"]:
            if f["name"] == old_name:
                f["name"] = name
                f["exec"] = exec
                f["directory"] = directory
                break
        else:
            abort(417, description="Unknown name='{}'".format(old_name))

        with open(cache["runner-path"], "w") as f:
            f.write(json.dumps(data))

        return {"status": 200}

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("name", type=str)
        parser.add_argument("exec", type=str)
        parser.add_argument("directory", type=str)
        args = parser.parse_args()
        name = args["name"]
        exec = args["exec"]
        directory = args["directory"]

        with open(cache["runner-path"]) as f:
            data = json.loads(f.read())

        for f in data["federates"]:
            if f["name"] == name:
                return abort(417, description="Name already exists {}".format(name))

        data["federates"].append(
            {
                "directory": directory,
                "exec": exec,
                "host": "localhost",
                "name": name,
            }
        )

        with open(cache["runner-path"], "w") as f:
            f.write(json.dumps(data))

        return {"status": 200}


api.add_resource(RunnerFileEdit, "/api/runner/file/edit")


class RunnerLog(Resource):
    def get(self):
        parser = reqparse.RequestParser()
        parser.add_argument("name", type=str)
        args = parser.parse_args()
        name = args["name"]
        with open(cache["runner-path"]) as f:
            data = json.loads(f.read())
        with open(os.path.join(os.path.dirname(cache["runner-path"]), "{}.log".format(name))) as f:
            data = f.read()
        return {"log": data}


api.add_resource(RunnerLog, "/api/runner/log")


class RunnerRun(Resource):
    runner_server = {}

    def get(self):
        if self.runner_server.get("process", None) is not None:
            return {"status": True}
        else:
            return {"status": False}

    def post(self):
        if self.get()["status"]:
            self.delete()
        p = subprocess.Popen(shlex.split("helics run --path {}".format(cache["runner-path"])))
        self.runner_server["process"] = p
        return {"status": True}

    def delete(self):
        self.runner_server["process"].terminate()
        self.runner_server["process"].kill()
        counter = 0
        while self.runner_server["process"].poll() is None or counter > 5:
            time.sleep(1)
            self.runner_server["process"].terminate()
            self.runner_server["process"].kill()
            counter += 1
        del self.runner_server["process"]
        global status_tracker
        status_tracker = {}
        return {"status": False}


api.add_resource(RunnerRun, "/api/runner/run")


class RunnerStatus(Resource):
    def get(self):
        parser = reqparse.RequestParser()
        parser.add_argument("name", type=str, required=True, help="Name of federate")
        args = parser.parse_args()
        return {"status": status_tracker[args["name"]]}

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("name", type=str, required=True, help="Name of federate")
        parser.add_argument("status", type=str, required=True, help="Status of federate")
        args = parser.parse_args()
        status_tracker[args["name"]] = args["status"]
        return {"status": status_tracker[args["name"]]}


api.add_resource(RunnerStatus, "/api/runner/status")


class Health(Resource):
    def get(self):
        return {"status": 200}


api.add_resource(Health, "/api/health")


class Profile(Resource):

    # <PROFILING>SenderFederate2[131074](initializing)HELICS CODE ENTRY<4570827706580384>[t=-1000000]</PROFILING>
    PATTERN = re.compile(
        r"""
                (?P<name>\w+)                           # SenderFederate2
                \[(\d+)\]                               # [131074]
                \((?P<state>\w+)\)                      # (initializing)
                (?P<message>(?:\w|\s)+)                 # HELICS CODE ENTRY
                \<(?P<realtime>\d+(?:\|\d+)?)\>         # <4570827706580384|534534523453>
                \[t=(?P<simtime>-?\d*\.{0,1}\d+)\]      # [t=-1000000]
                """,
        re.X,
    )

    def get(self):
        with open(cache["profile-path"]) as f:
            data = f.read()
        data = data.replace(r"<PROFILING>", "").replace(r"</PROFILING>", "")
        names = []
        states = []
        messages = []
        simtimes = []
        realtimes = []
        time_marker = {}
        for line in data.splitlines():
            m = self.PATTERN.match(line)
            m = cast(re.Match[str], m)
            name = m.group("name")
            state = m.group("state")
            message = m.group("message")
            simtime = float(m.group("simtime"))
            try:
                realtime = float(m.group("realtime"))
            except ValueError:
                realtime, markertime = m.group("realtime").split("|")
                time_marker[name] = float(markertime)
                realtime = float(realtime)
            names.append(name)
            states.append(state)
            messages.append(message)
            simtimes.append(simtime)
            realtimes.append(realtime)

        profile = {}
        for name in set(names):
            profile[name] = []

        invert = True
        if invert:
            for name in set(names):
                profile[name].append({})

        for (name, state, message, simtime, realtime) in zip(names, states, messages, simtimes, realtimes):
            if state == "created":
                continue
            if "ENTRY" in message and not invert:
                profile[name].append({"s_enter": simtime, "r_enter": realtime})
            elif "EXIT" in message and not invert:
                profile[name][-1]["s_end"] = simtime
                profile[name][-1]["r_end"] = realtime
            elif "EXIT" in message and invert:
                profile[name].append({"s_enter": simtime, "r_enter": realtime})
            elif "ENTRY" in message and invert:
                profile[name][-1]["s_end"] = simtime
                profile[name][-1]["r_end"] = realtime
        profiles = {}
        names = {k: i for i, k in enumerate(sorted(profile.keys()))}

        end = "r_end"
        enter = "r_enter"
        scaling = 1e9

        for k in profile.keys():
            profiles[k] = []
            for i in profile[k]:
                if end in i.keys() and enter in i.keys():
                    i["name"] = k
                    i[enter] = i[enter] / scaling
                    i[end] = i[end] / scaling
                    profiles[k].append(i)

        return profiles

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("file", type=werkzeug.datastructures.FileStorage, location="files")
        args = parser.parse_args()
        file = args["file"]
        os.makedirs(app.config["UPLOAD_FOLDER"], exist_ok=True)
        file.save(os.path.join(app.config["UPLOAD_FOLDER"], "profile.txt"))
        cache["profile-path"] = os.path.join(app.config["UPLOAD_FOLDER"], "profile.txt")


api.add_resource(Profile, "/api/profiler/")


class APIException(Exception):
    def __init__(self, code, message):
        self._code = code
        self._message = message

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    def __str__(self):
        return self.__class__.__name__ + ": " + self.message


class BrokerServer(Resource):
    broker_server = {}

    def get(self):
        if self.broker_server.get("process", None) is not None:
            return {"status": True}
        else:
            return {"status": False}

    def post(self):
        parser = reqparse.RequestParser()
        parser.add_argument("status", type=bool, required=True, help="requested status of broker server")
        args = parser.parse_args()
        status = args["status"]
        if status is True and self.broker_server.get("process", None) is not None:
            return abort(417, description="Unable to start server", status=True)
        elif status is False and self.broker_server.get("process", None) is None:
            return abort(417, description="Unable to stop server", status=False)
        elif status is True:
            p = subprocess.Popen(shlex.split("helics_broker_server --http"))
            self.broker_server["process"] = p
            return {"status": status}
        elif status is False:
            self.broker_server["process"].terminate()
            self.broker_server["process"].kill()
            counter = 0
            # TODO: Find out why broker server is not being terminated
            while self.broker_server["process"].poll() is None or counter > 5:
                time.sleep(1)
                self.broker_server["process"].terminate()
                self.broker_server["process"].kill()
                counter += 1
            del self.broker_server["process"]
            return {"status": status}


api.add_resource(BrokerServer, "/api/broker-server")


@app.route("/", defaults={"path": "index.html"})
@app.route("/<path:path>")
def index(path):
    return send_from_directory(os.path.join(current_directory, "..", "static"), path)


def run():
    debug = bool(os.environ.get("PYHELICS_FLASK_DEBUG", False))
    if debug:
        host = None
    else:
        host = "0.0.0.0"
        cli = sys.modules["flask.cli"]
        cli.show_server_banner = lambda *x: None
        # os.environ["WERKZEUG_RUN_MAIN"] = "true"
    app.run(host=host, debug=debug)



# -*- coding: utf-8 -*-
import cffi
import os
import platform
import warnings
import logging

ffi = cffi.FFI()

logger = logging.getLogger(__name__)


CURRENT_INSTALL = os.path.join(os.path.dirname(os.path.abspath(__file__)), "install")
HELICS_INSTALL = os.getenv(
    "HELICS_INSTALL",
    CURRENT_INSTALL,
)
PYHELICS_INSTALL = os.getenv("PYHELICS_INSTALL", HELICS_INSTALL)
if not os.path.isdir(PYHELICS_INSTALL):
    warnings.warn("PYHELICS_INSTALL ({}) is not a directory. Using DEFAULT_INSTALL ({})".format(PYHELICS_INSTALL, CURRENT_INSTALL))
    PYHELICS_INSTALL = CURRENT_INSTALL

files = [
    "helics_api.h",
    "helics_enums.h",
    os.path.join("shared_api_library", "api-data.h"),
    os.path.join("shared_api_library", "helics.h"),
    os.path.join("shared_api_library", "helics_export.h"),
    os.path.join("shared_api_library", "MessageFederate.h"),
    os.path.join("shared_api_library", "MessageFilters.h"),
    os.path.join("shared_api_library", "ValueFederate.h"),
    os.path.join("shared_api_library", "helicsCallbacks.h"),
]
IGNOREBLOCK = False
for file in files:
    if not os.path.isfile(os.path.join(PYHELICS_INSTALL, "include", "helics", file)):
        continue
    with open(os.path.join(PYHELICS_INSTALL, "include", "helics", file)) as f:
        lines = []
        for line in f:
            if line.startswith("#ifdef __cplusplus"):
                IGNOREBLOCK = True
                continue
            if IGNOREBLOCK is True and line.startswith("#endif"):
                IGNOREBLOCK = False
                continue
            if IGNOREBLOCK is True:
                continue
            if line.startswith("#"):
                continue
            lines.append(line)
        data = "\n".join(lines)
        data = data.replace("HELICS_EXPORT", "")
        data = data.replace("HELICS_DEPRECATED_EXPORT", "")
        ffi.cdef(data)
        if file.endswith("helics_api.h"):
            break

# ffi.set_source(
#     "_py_helics",
#     f"""#include "{CHELICS}" """,
#     libraries=["helicsSharedLib"],
#     library_dirs=[os.path.join(PYHELICS_INSTALL, "lib")],
# )


def _load_library():
    lib = None
    if platform.system() == "Windows":
        logger.debug("Loading helics library for windows.")
        for file in os.listdir(os.path.join(PYHELICS_INSTALL, "bin")):
            if "helics" in file and file.endswith(".dll"):
                try:
                    logger.debug("dlopen helics library in {}".format(os.path.join(PYHELICS_INSTALL, "bin", file)))
                    lib = ffi.dlopen(os.path.join(PYHELICS_INSTALL, "bin", file))
                    break
                except OSError as _:
                    pass
        else:
            try:
                try:
                    logger.debug("dlopen system helics library helics.dll")
                    lib = ffi.dlopen("helics.dll")
                except:
                    logger.debug("dlopen system helics library helicsShared.dll")
                    lib = ffi.dlopen("helicsShared.dll")
            except OSError as e:
                raise OSError(
                    str(e)
                    + "\n\nRECOMMENDATION: When using Python / Anaconda on Windows, users must manually install the latest version of Visual C++ Redistributable for Visual Studio 2019. See https://support.microsoft.com/en-us/help/2977003/the-latest-supported-visual-c-downloads for links. If this problem persists after installing Visual C++ Redistributable, please open an issue on https://github.com/GMLC-TDC/HELICS."
                )
            if lib is None:
                raise Exception("Unable to load helics shared library")
    elif platform.system() == "Darwin":
        logger.debug("Loading helics library for macos.")
        if os.path.isdir(os.path.join(PYHELICS_INSTALL, "lib64")):
            lib_folder = os.path.join(PYHELICS_INSTALL, "lib64")
            logger.debug("lib64 folder found.")
        else:
            lib_folder = os.path.join(PYHELICS_INSTALL, "lib")
            logger.debug("lib folder found.")
        for file in reversed(sorted(os.listdir(lib_folder), key=len)):
            if "helicsSharedLib." in file or "libhelics." in file and file.endswith(".dylib"):
                logger.debug("dlopen helics library in {}".format(os.path.join(lib_folder, file)))
                lib = ffi.dlopen(os.path.join(lib_folder, file))
                break
        else:
            for file in reversed(sorted(os.listdir(lib_folder), key=len)):
                if "helicsSharedLibd." in file or "libhelicsd." in file and file.endswith(".dylib"):
                    logger.debug("dlopen debug helics library in {}".format(os.path.join(lib_folder, file)))
                    lib = ffi.dlopen(os.path.join(lib_folder, file))
                    break
            else:
                try:
                    logger.debug("dlopen system helics library libhelics.dylib")
                    lib = ffi.dlopen("libhelics.dylib")
                except:
                    logger.debug("dlopen system helics library helicsSharedLib.dylib")
                    lib = ffi.dlopen("helicsSharedLib.dylib")
                if lib is None:
                    raise Exception("Unable to load helics shared library")
    elif platform.system() == "Linux":
        logger.debug("Loading helics library for linux.")
        if os.path.isdir(os.path.join(PYHELICS_INSTALL, "lib64")):
            lib_folder = os.path.join(PYHELICS_INSTALL, "lib64")
            logger.debug("lib64 folder found.")
        else:
            lib_folder = os.path.join(PYHELICS_INSTALL, "lib")
            logger.debug("lib folder found.")
        for file in reversed(sorted(os.listdir(lib_folder), key=len)):
            if "helicsSharedLib." in file or "libhelics." in file and file.endswith(".so"):
                logger.debug("dlopen helics library in {}".format(os.path.join(lib_folder, file)))
                lib = ffi.dlopen(os.path.join(lib_folder, file))
                break
        else:
            for file in reversed(sorted(os.listdir(lib_folder), key=len)):
                if "helicsSharedLibd." in file or "libhelicsd." in file and file.endswith(".so"):
                    try:
                        logger.debug("dlopen debug helics library in {}".format(os.path.join(lib_folder, file)))
                        lib = ffi.dlopen(os.path.join(lib_folder, file))
                        break
                    except:
                        pass
            else:
                try:
                    logger.debug("dlopen system helics library libhelics.so")
                    lib = ffi.dlopen("libhelics.so")
                except:
                    logger.debug("dlopen system helics library helicsSharedLib.so")
                    lib = ffi.dlopen("helicsSharedLib.so")
                if lib is None:
                    raise Exception("Unable to load helics shared library")
    return lib


lib = _load_library()



# -*- coding: utf-8 -*-
import os
import sys

CURRENT_DIRECTORY = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(CURRENT_DIRECTORY)
sys.path.append(os.path.dirname(CURRENT_DIRECTORY))

import time
import helics as h
import os
import pytest as pt

from test_init import createBroker, createValueFederate, destroyFederate, destroyBroker, createMessageFederate


CURRENT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))


def test_valuefederate_creation():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()
    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_state():

    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    state = h.helicsFederateGetState(vFed)
    assert state == 0

    h.helicsFederateEnterExecutingMode(vFed)

    state = h.helicsFederateGetState(vFed)
    assert state == 2

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_publication_registration():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    pubid1 = h.helicsFederateRegisterTypePublication(vFed, "pub1", "string", "")
    pubid2 = h.helicsFederateRegisterGlobalTypePublication(vFed, "pub2", "int", "")
    pubid3 = h.helicsFederateRegisterTypePublication(vFed, "pub3", "double", "V")

    h.helicsFederateEnterExecutingMode(vFed)

    assert h.helicsPublicationGetName(pubid1) == "TestA Federate/pub1"
    assert h.helicsPublicationGetName(pubid2) == "pub2"

    assert h.helicsPublicationGetName(pubid3) == "TestA Federate/pub3"
    assert h.helicsPublicationGetType(pubid3) == "double"
    assert h.helicsPublicationGetUnits(pubid3) == "V"

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_named_point():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    defaultValue = "start of a longer string in place of the shorter one and now this should be very long"
    defVal = 5.3
    # testValue1 = "inside of the functional relationship of helics"
    testValue1 = "short string"
    testVal1 = 45.7823
    testValue2 = "I am a string"
    testVal2 = 0.0

    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_NAMED_POINT, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")

    h.helicsInputSetDefaultNamedPoint(subid, defaultValue, defVal)

    h.helicsFederateEnterExecutingMode(vFed)

    # publish string1 at time=0.0
    h.helicsPublicationPublishNamedPoint(pubid, testValue1, testVal1)

    assert h.helicsInputGetNamedPoint(subid) == (defaultValue, defVal)

    assert h.helicsFederateRequestTime(vFed, 1.0) == 0.01

    # # get the value
    assert h.helicsInputGetNamedPoint(subid) == (testValue1, testVal1)

    # publish a second string
    h.helicsPublicationPublishNamedPoint(pubid, testValue2, testVal2)

    # # make sure the value is still what we expect
    assert h.helicsInputGetNamedPoint(subid) == (testValue1, testVal1)
    # # make sure the string is what we expect
    # # assert value3 == testValue1
    # assert val3 == [testValue1, testVal1]

    # # advance time
    assert h.helicsFederateRequestTime(vFed, 2.0) == 0.02

    # # make sure the value was updated
    assert h.helicsInputGetNamedPoint(subid) == (testValue2, testVal2)

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_bool():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    defaultValue = True
    testValue1 = True
    testValue2 = False

    # register the publications
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_BOOLEAN, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")

    h.helicsInputSetDefaultBoolean(subid, defaultValue)

    h.helicsFederateEnterExecutingMode(vFed)

    # publish string1 at time=0.0
    h.helicsPublicationPublishBoolean(pubid, testValue1)
    val = h.helicsInputGetBoolean(subid)

    assert val == defaultValue

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    # get the value
    val = h.helicsInputGetBoolean(subid)

    # make sure the string is what we expect
    assert val == testValue1

    # publish a second string
    h.helicsPublicationPublishBoolean(pubid, testValue2)

    # make sure the value is still what we expect
    val = h.helicsInputGetBoolean(subid)
    assert val == testValue1
    # advance time
    grantedtime = h.helicsFederateRequestTime(vFed, 2.0)
    # make sure the value was updated
    assert grantedtime == 0.02

    val = h.helicsInputGetBoolean(subid)
    assert val == testValue2

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_publisher_registration():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    pubid1 = h.helicsFederateRegisterPublication(vFed, "pub1", h.HELICS_DATA_TYPE_STRING, "")
    pubid2 = h.helicsFederateRegisterGlobalPublication(vFed, "pub2", h.HELICS_DATA_TYPE_INT, "")
    pubid3 = h.helicsFederateRegisterPublication(vFed, "pub3", h.HELICS_DATA_TYPE_DOUBLE, "V")
    h.helicsFederateEnterExecutingMode(vFed)

    publication_key = h.helicsPublicationGetName(pubid1)
    assert publication_key == "TestA Federate/pub1"
    publication_type = h.helicsPublicationGetType(pubid1)
    assert publication_type == "string"
    publication_key = h.helicsPublicationGetName(pubid2)
    assert publication_key == "pub2"
    publication_key = h.helicsPublicationGetName(pubid3)
    assert publication_key == "TestA Federate/pub3"
    publication_type = h.helicsPublicationGetType(pubid3)
    assert publication_type == "double"
    publication_units = h.helicsPublicationGetUnits(pubid3)
    assert publication_units == "V"
    publication_type = h.helicsPublicationGetType(pubid2)
    assert publication_type == "int64"

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_subscription_and_publication_registration():

    broker = createBroker()
    vFed, fedinfo = createValueFederate(1, "fed0")

    pubid = h.helicsFederateRegisterPublication(vFed, "pub1", h.HELICS_DATA_TYPE_STRING, "")
    pubid2 = h.helicsFederateRegisterGlobalPublication(vFed, "pub2", h.HELICS_DATA_TYPE_INT, "volts")
    pubid3 = h.helicsFederateRegisterTypePublication(vFed, "pub3", "double", "V")

    subid1 = h.helicsFederateRegisterSubscription(vFed, "sub1", "")
    subid2 = h.helicsFederateRegisterSubscription(vFed, "sub2", "")
    subid3 = h.helicsFederateRegisterSubscription(vFed, "sub3", "V")

    h.helicsFederateEnterExecutingMode(vFed)

    publication_type = h.helicsPublicationGetType(pubid3)
    assert publication_type == "double"

    sub_key = h.helicsSubscriptionGetTarget(subid1)
    assert sub_key == "sub1"
    sub_type = h.helicsInputGetType(subid1)
    assert sub_type == ""
    sub_key = h.helicsSubscriptionGetTarget(subid2)
    assert sub_key == "sub2"
    sub_key = h.helicsSubscriptionGetTarget(subid3)
    assert sub_key == "sub3"
    sub_type = h.helicsInputGetType(subid3)
    assert sub_type == ""
    sub_units = h.helicsInputGetUnits(subid3)
    assert sub_units == "V"
    sub_type = h.helicsInputGetType(subid2)
    assert sub_type == ""

    subid_b = h.helicsFederateGetSubscription(vFed, "sub1")
    tmp = h.helicsSubscriptionGetTarget(subid_b)
    assert tmp == "sub1"
    # check the getSubscriptionByIndex function
    subid_c = h.helicsFederateGetInputByIndex(vFed, 2)
    tmp = h.helicsInputGetUnits(subid_c)
    assert tmp == "V"
    # check publications

    sv = h.helicsPublicationGetName(pubid)
    sv2 = h.helicsPublicationGetName(pubid2)
    assert sv == "Testfed0/pub1"
    assert sv2 == "pub2"
    pub3name = h.helicsPublicationGetName(pubid3)
    assert pub3name == "Testfed0/pub3"

    type = h.helicsPublicationGetType(pubid3)
    assert type == "double"
    units = h.helicsPublicationGetUnits(pubid3)
    assert units == "V"

    # check the getSubscription function

    pubid_b = h.helicsFederateGetPublication(vFed, "Testfed0/pub1")
    tmp = h.helicsPublicationGetType(pubid_b)
    assert tmp == "string"
    # check the getSubscriptionByIndex function
    pubid_c = h.helicsFederateGetPublicationByIndex(vFed, 1)
    tmp = h.helicsPublicationGetUnits(pubid_c)
    assert tmp == "volts"

    # this one should be invalid
    # @test_throws h.HELICSErrorInvalidArgument pubid_d = h.helicsFederateGetPublicationByIndex(vFed, 5)

    h.helicsFederateDisconnect(vFed)

    state = h.helicsFederateGetState(vFed)
    assert state == h.HELICS_STATE_FINALIZE

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_single_transfer():

    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_STRING, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")

    h.helicsFederateEnterExecutingMode(vFed)

    h.helicsPublicationPublishString(pubid, "string1")

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    s = h.helicsInputGetString(subid)
    assert s == "string1"

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_double():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    defaultValue = 1.0
    testValue = 2.0
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_DOUBLE, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")
    h.helicsInputSetDefaultDouble(subid, defaultValue)

    h.helicsFederateEnterExecutingMode(vFed)

    # publish string1 at time=0.0
    h.helicsPublicationPublishDouble(pubid, testValue)

    value = h.helicsInputGetDouble(subid)
    assert value == defaultValue

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    value = h.helicsInputGetDouble(subid)
    assert value == testValue

    # publish string1 at time=0.0
    h.helicsPublicationPublishDouble(pubid, testValue + 1)

    grantedtime = h.helicsFederateRequestTime(vFed, 2.0)
    assert grantedtime == 0.02

    value = h.helicsInputGetDouble(subid)
    assert value == testValue + 1

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_complex():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    rDefaultValue = 1.0
    iDefaultValue = 1.0
    rTestValue = 2.0
    iTestValue = 2.0
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_COMPLEX, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")
    h.helicsInputSetDefaultComplex(subid, complex(rDefaultValue, iDefaultValue))

    h.helicsFederateEnterExecutingMode(vFed)

    # publish string1 at time=0.0
    h.helicsPublicationPublishComplex(pubid, complex(rTestValue, iTestValue))

    assert complex(rDefaultValue, iDefaultValue) == h.helicsInputGetComplex(subid)

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    assert complex(rTestValue, iTestValue) == h.helicsInputGetComplex(subid)

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_integer():

    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    defaultValue = 1
    testValue = 2
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_INT, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")
    h.helicsInputSetDefaultInteger(subid, defaultValue)

    h.helicsFederateEnterExecutingMode(vFed)

    h.helicsPublicationPublishInteger(pubid, testValue)

    value = h.helicsInputGetInteger(subid)
    assert value == defaultValue

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    value = h.helicsInputGetInteger(subid)
    assert value == testValue

    h.helicsPublicationPublishInteger(pubid, testValue + 1)

    grantedtime = h.helicsFederateRequestTime(vFed, 2.0)
    assert grantedtime == 0.02

    value = h.helicsInputGetInteger(subid)
    assert value == testValue + 1

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_string():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    defaultValue = "String1"
    testValue = "String2"
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_STRING, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")
    h.helicsInputSetDefaultString(subid, defaultValue)

    h.helicsFederateEnterExecutingMode(vFed)

    h.helicsPublicationPublishString(pubid, testValue)

    value = h.helicsInputGetString(subid)
    assert value == defaultValue

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    value = h.helicsInputGetString(subid)
    assert value == testValue

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_vectord():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    defaultValue = [0.0, 1.0, 2.0]
    testValue = [3.0, 4.0, 5.0]
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_VECTOR, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")
    h.helicsInputSetDefaultVector(subid, defaultValue)

    h.helicsFederateEnterExecutingMode(vFed)

    h.helicsPublicationPublishVector(pubid, testValue)

    assert h.helicsInputGetVector(subid) == defaultValue

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    value = h.helicsInputGetVector(subid)

    assert value == testValue

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_single_transfer():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()

    s = "n2"

    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_STRING, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")

    h.helicsFederateEnterExecutingMode(vFed)

    h.helicsPublicationPublishString(pubid, "string1")

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    s = h.helicsInputGetString(subid)

    assert s == "string1"

    t = h.helicsInputLastUpdateTime(subid)
    assert t == 0.01

    h.helicsPublicationPublishString(pubid, "string2")

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_default_value_tests():
    broker = createBroker()
    vFed1, fedinfo = createValueFederate(1, "fed0")

    inp_raw1 = h.helicsFederateRegisterInput(vFed1, "key1", h.HELICS_DATA_TYPE_RAW, "raw")
    inp_raw2 = h.helicsFederateRegisterInput(vFed1, "key2", h.HELICS_DATA_TYPE_RAW, "raw")

    inp_bool = h.helicsFederateRegisterInput(vFed1, "key3", h.HELICS_DATA_TYPE_BOOLEAN, "")

    inp_time = h.helicsFederateRegisterInput(vFed1, "key4", h.HELICS_DATA_TYPE_TIME, "")

    inp_char = h.helicsFederateRegisterInput(vFed1, "key5", h.HELICS_DATA_TYPE_STRING, "")

    inp_vect = h.helicsFederateRegisterInput(vFed1, "key6", h.HELICS_DATA_TYPE_VECTOR, "V")

    inp_double = h.helicsFederateRegisterInput(vFed1, "key7", h.HELICS_DATA_TYPE_DOUBLE, "kW")

    inp_double2 = h.helicsFederateRegisterInput(vFed1, "key8", h.HELICS_DATA_TYPE_DOUBLE, "")

    inp_np = h.helicsFederateRegisterInput(vFed1, "key9", h.HELICS_DATA_TYPE_NAMED_POINT, "")

    h.helicsInputSetMinimumChange(inp_double, 1100.0)
    h.helicsInputSetDefaultDouble(inp_double, 10000.0)

    h.helicsInputSetOption(inp_double2, h.HELICS_HANDLE_OPTION_CONNECTION_REQUIRED, True)

    pub = h.helicsFederateRegisterPublication(vFed1, "", h.HELICS_DATA_TYPE_INT, "MW")
    h.helicsPublicationSetOption(pub, h.HELICS_HANDLE_OPTION_CONNECTION_REQUIRED, True)
    h.helicsPublicationAddTarget(pub, "Testfed0/key7")
    h.helicsPublicationAddTarget(pub, "Testfed0/key8")

    h.helicsInputSetDefaultBytes(inp_raw1, "")
    data = "this is a string"
    h.helicsInputSetDefaultBytes(inp_raw2, data)

    h.helicsInputSetDefaultBoolean(inp_bool, True)

    h.helicsInputSetDefaultTime(inp_time, 12.3)
    h.helicsInputSetDefaultChar(inp_char, "q")
    h.helicsInputSetDefaultVector(inp_vect, [])
    h.helicsInputSetDefaultNamedPoint(inp_np, data, 15.7)

    h.helicsFederateEnterExecutingMode(vFed1)
    assert h.helicsInputGetInjectionUnits(inp_double) == "MW"
    assert h.helicsInputGetInjectionUnits(inp_double2) == "MW"
    assert h.helicsInputGetType(inp_double) == "double"
    assert h.helicsInputGetPublicationType(inp_double) == "int64"

    c2 = h.helicsInputGetChar(inp_char)
    assert c2 == "q"
    h.helicsInputGetVector(inp_vect)

    optset = h.helicsInputGetOption(inp_double2, h.HELICS_HANDLE_OPTION_CONNECTION_REQUIRED)
    assert optset == 1

    optset = h.helicsPublicationGetOption(pub, h.HELICS_HANDLE_OPTION_CONNECTION_REQUIRED)
    assert optset == 1
    h.helicsPublicationPublishInteger(pub, 12)

    h.helicsFederateRequestNextStep(vFed1)
    assert h.helicsInputGetDouble(inp_double) == 12000.0
    assert h.helicsInputGetDouble(inp_double2) == 12.0

    h.helicsPublicationPublishInteger(pub, 13)

    h.helicsFederateRequestNextStep(vFed1)
    assert h.helicsInputIsUpdated(inp_double) is False
    assert h.helicsInputIsUpdated(inp_double2) is True

    assert h.helicsInputGetDouble(inp_double) == 12000.0
    assert h.helicsInputGetDouble(inp_double2) == 13.0

    h.helicsPublicationPublishInteger(pub, 15)

    h.helicsFederateRequestNextStep(vFed1)

    assert h.helicsInputIsUpdated(inp_double) is True
    assert h.helicsInputIsUpdated(inp_double2) is True

    h.helicsInputClearUpdate(inp_double)
    h.helicsInputClearUpdate(inp_double2)

    assert h.helicsInputIsUpdated(inp_double) is False
    assert h.helicsInputIsUpdated(inp_double2) is False

    _, rval = h.helicsInputGetNamedPoint(inp_np)
    assert rval == 15.7

    out, rval = h.helicsInputGetNamedPoint(inp_np)
    assert out == "this is a string"
    assert rval == 15.7

    h.helicsFederateDisconnect(vFed1)

    destroyFederate(vFed1, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_info_filed():

    broker = createBroker()
    vFed, fedinfo = createValueFederate(1, "fed0")

    h.helicsFederateSetFlagOption(vFed, h.HELICS_HANDLE_OPTION_CONNECTION_OPTIONAL, True)
    # register the publications/subscriptions

    subid1 = h.helicsFederateRegisterSubscription(vFed, "sub1", "")
    pubid1 = h.helicsFederateRegisterTypePublication(vFed, "pub1", "string", "")
    pubid2 = h.helicsFederateRegisterGlobalTypePublication(vFed, "pub2", "string", "")

    # Set info fields
    h.helicsInputSetInfo(subid1, "sub1_test")
    h.helicsPublicationSetInfo(pubid1, "pub1_test")
    h.helicsPublicationSetInfo(pubid2, "pub2_test")
    h.helicsFederateEnterExecutingMode(vFed)

    assert h.helicsInputGetInfo(subid1) == "sub1_test"
    assert h.helicsPublicationGetInfo(pubid1) == "pub1_test"
    assert h.helicsPublicationGetInfo(pubid2) == "pub2_test"

    cr = h.helicsFederateGetCore(vFed)
    h.helicsFederateDisconnect(vFed)

    wait = h.helicsCoreWaitForDisconnect(cr, 70)
    if wait is False:
        wait = h.helicsCoreWaitForDisconnect(cr, 500)
    assert wait is True

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)


def test_valuefederate_test_file_load():

    filename = os.path.join(CURRENT_DIRECTORY, "valuefederate.json")
    vFed = h.helicsCreateValueFederateFromConfig(filename)

    name = h.helicsFederateGetName(vFed)
    assert name == "valueFed"

    assert h.helicsFederateGetInputCount(vFed) == 3
    assert h.helicsFederateGetPublicationCount(vFed) == 2

    h.helicsFederateDisconnect(vFed)
    h.helicsFederateFree(vFed)
    h.helicsCloseLibrary()


def test_valuefederate_test_bytes():
    broker = createBroker()
    vFed, fedinfo = createValueFederate()
    defaultValue = bytes([1, 2, 3, 4, 5])
    testValue = bytes([5, 4, 3, 2, 1, 0, 255, 30, 17, 18, 19])
    pubid = h.helicsFederateRegisterGlobalPublication(vFed, "pub1", h.HELICS_DATA_TYPE_RAW, "")
    subid = h.helicsFederateRegisterSubscription(vFed, "pub1", "")
    h.helicsInputSetDefaultBytes(subid, defaultValue)

    h.helicsFederateEnterExecutingMode(vFed)

    h.helicsPublicationPublishBytes(pubid, testValue)

    value = h.helicsInputGetBytes(subid)
    assert value == defaultValue

    grantedtime = h.helicsFederateRequestTime(vFed, 1.0)
    assert grantedtime == 0.01

    value = h.helicsInputGetBytes(subid)
    assert value == testValue

    destroyFederate(vFed, fedinfo)
    destroyBroker(broker)



# -*- coding: utf-8 -*-
from sqlalchemy import Column, Integer, Float, String, JSON, Sequence, Table, Boolean, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

Base = declarative_base()


def as_dict(t):
    return {c.name: getattr(t, c.name) for c in t.__table__.columns}


class MetaData(Base):
    __tablename__ = "metadata"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    name = Column(String)
    value = Column(String)


class SystemInfo(Base):
    __tablename__ = "systeminfo"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    data = Column(JSON)


class FederateGraph(Base):
    __tablename__ = "federategraph"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    data = Column(JSON)


class DataGraph(Base):
    __tablename__ = "datagraph"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    data = Column(JSON)


class Cores(Base):
    __tablename__ = "cores"
    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(String)
    parent = Column(Integer)
    address = Column(String)


class Federates(Base):
    __tablename__ = "federates"
    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(String)
    parent = Column(Integer)


class FederateEventLogs(Base):
    __tablename__ = "federateeventlogs"
    id = Column(Integer, primary_key=True, nullable=False)
    updated_at = Column(Float)
    name = Column(String)
    requested_time = Column(Float)
    granted_time = Column(Float)
    state = Column(String)


class Brokers(Base):
    __tablename__ = "brokers"
    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(String)


class Publications(Base):
    __tablename__ = "publications"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    federate_name = Column(String)
    name = Column(String)
    source = Column(Integer)
    target = Column(Integer)
    units = Column(String)
    type = Column(String)
    info = Column(String)
    is_valid = Column(Boolean)


class Subscriptions(Base):
    __tablename__ = "subscriptions"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    federate_name = Column(String)
    name = Column(String)
    publication_name = Column(String)
    source = Column(Integer)
    target = Column(Integer)
    units = Column(String)
    type = Column(String)
    info = Column(String)
    injection_units = Column(String)
    is_valid = Column(Boolean)


class Inputs(Base):
    __tablename__ = "inputs"
    id = Column(Integer, Sequence("id"), primary_key=True, nullable=False)
    federate_name = Column(String)
    publication_name = Column(String)
    source = Column(Integer)
    target = Column(Integer)
    units = Column(String)
    type = Column(String)
    info = Column(String)
    injection_units = Column(String)
    is_valid = Column(Boolean)



# -*- coding: utf-8 -*-
import threading
import time
import logging
import json
import urllib.request

from .utils import error


class HELICSRuntimeError(RuntimeError):
    pass


logger = logging.getLogger(__name__)

API = "http://127.0.0.1:5000/api"


class CheckStatusThread(threading.Thread):
    def __init__(self, process_list, should_kill, helics_server_available):
        threading.Thread.__init__(self)
        self.should_kill = should_kill
        self._process_list = process_list
        self._status = {}
        self._helics_server_available = helics_server_available

    def run(self):
        logger.info("Starting checks logger")
        for p in self._process_list:
            self._status[p.name] = 0
        has_failed = False
        while True:
            time.sleep(1)
            for p in self._process_list:
                self.status(p)
                if has_failed is True and self.should_kill is True:
                    p.process.kill()
                if p.process.poll() is not None and p.process.returncode != 0:
                    self._status[p.name] = p.process.returncode
                    if has_failed is False and self.should_kill is True:
                        error("Process {} has failed, killing other processes".format(p.name))
                    has_failed = True

            all_p = [p.process.poll() for p in self._process_list if p.process.poll() is not None]
            if len(all_p) == len(self._process_list):
                if all(p == 0 for p in all_p):
                    return 0
                else:
                    return -1
            else:
                continue

    def status(self, p):
        if self._helics_server_available:
            if p.process.poll() is None:
                status = "running"
            elif p.process.returncode == 0:
                status = "success"
            elif p.process.returncode == -9:
                status = "terminated"
            elif p.process.returncode != 0:
                status = "failed"
            else:
                status = "unknown"
            r = urllib.request.Request("{}/runner/status".format(API))
            r.add_header("Content-Type", "application/json; charset=utf-8")
            body = {"name": p.name, "status": status}
            bytes = json.dumps(body).encode("utf-8")
            r.add_header("Content-Length", str(len(bytes)))
            try:
                with urllib.request.urlopen(r, bytes) as response:
                    json.loads(response.read().decode(response.info().get_param("charset") or "utf-8"))
            except Exception as e:
                logger.exception("Severing connection with helics-cli server: {}".format(e))
                self._helics_server_available = False



# -*- coding: utf-8 -*-
import os
import sys

CURRENT_DIRECTORY = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(CURRENT_DIRECTORY)
sys.path.append(os.path.dirname(CURRENT_DIRECTORY))

import time
import helics as h


def createBroker(number=1):
    initstring = f"-f {number} --name=mainbroker"
    # @test_throws h.HELICSErrorInvalidArgument broker = h.helicsCreateBroker("mq", "", initstring)
    broker = h.helicsCreateBroker("zmq", "", initstring)
    # assert broker is h.Broker
    assert h.helicsBrokerIsConnected(broker) is True
    return broker


def setupFederateInfo(name="A Core", number=1, deltat=0.01):
    fedinitstring = f"--broker=mainbroker --federates={number}"

    # Create Federate Info object that describes the federate properties
    fedinfo = h.helicsCreateFederateInfo()
    # assert fedinfo isa h.FederateInfo

    # # Set Federate name
    h.helicsFederateInfoSetCoreName(fedinfo, f"Test{name}")

    # # Set core type from string
    h.helicsFederateInfoSetCoreTypeFromString(fedinfo, "zmq")

    # # Federate init string
    h.helicsFederateInfoSetCoreInitString(fedinfo, fedinitstring)

    # # Set the message interval (timedelta) for federate. Note th#
    # # HELICS minimum message time interval is 1 ns and by default
    # # it uses a time delta of 1 second. What is provided to the
    # # setTimedelta routine is a multiplier for the default timedelta.

    # # Set one second message interval
    h.helicsFederateInfoSetTimeProperty(fedinfo, h.HELICS_PROPERTY_TIME_DELTA, deltat)
    h.helicsFederateInfoSetIntegerProperty(fedinfo, h.HELICS_PROPERTY_INT_LOG_LEVEL, -1)
    return fedinfo


def createValueFederate(federates=1, name="A Federate", deltat=0.01):
    fedinfo = setupFederateInfo(name, federates, deltat)
    vFed = h.helicsCreateValueFederate(f"Test{name}", fedinfo)
    # assert vFed isa h.ValueFederate
    return vFed, fedinfo


def createMessageFederate(federates=1, name="A Federate", deltat=0.01):
    fedinfo = setupFederateInfo(name, federates, deltat)
    mFed = h.helicsCreateMessageFederate(f"Test{name}", fedinfo)
    # assert mFed isa h.MessageFederate
    return mFed, fedinfo


def destroyBroker(broker):
    h.helicsBrokerDisconnect(broker)
    h.helicsCloseLibrary()


def destroyFederate(fed, fedinfo, broker=None):
    h.helicsFederateDisconnect(fed)
    _ = h.helicsFederateGetState(fed)
    if broker is not None:
        while h.helicsBrokerIsConnected(broker):
            time.sleep(1)
    h.helicsFederateInfoFree(fedinfo)
    h.helicsFederateFree(fed)
    if broker is not None:
        destroyBroker(broker)


destroyValueFederate = destroyFederate
destroyMessageFederate = destroyFederate



# -*- coding: utf-8 -*-

import os
import sys

CURRENT_DIRECTORY = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(CURRENT_DIRECTORY)
sys.path.append(os.path.dirname(CURRENT_DIRECTORY))

import helics as h


def test_helicsGetVersion():
    print(h.helicsGetVersion())
    try:
        assert h.helicsGetVersion().startswith("2")
    except Exception:
        assert h.helicsGetVersion().startswith("3")



# -*- coding: utf-8 -*-

import logging
import os
import time
import tempfile
from typing import Dict, List, cast
from datetime import datetime

import helics as h

from . import database as db

logger = logging.getLogger(__name__)
hdlr = logging.StreamHandler()
fmt = logging.Formatter("%(asctime)s %(name)-35s %(levelname)-8s %(message)s")
hdlr.setFormatter(fmt)
logger.addHandler(hdlr)
logger.setLevel(logging.DEBUG)


class HelicsObserverFederate:
    """
    HelicsObserverFederate
    """

    def __init__(self, folder=None, core_type="zmq"):
        if folder is None:
            self._folder = tempfile.gettempdir()
        else:
            self._folder = folder
        self._core_type = core_type
        self._setup_engine()
        self._setup_federate()

    def _setup_engine(self):
        db_file = os.path.abspath(os.path.join(self._folder, "helics-cli.sqlite.db"))
        if os.path.exists(db_file):
            os.remove(db_file)
        self.engine = db.create_engine("sqlite+pysqlite:///{}".format(db_file), echo=False, future=True)
        db.Base.metadata.create_all(self.engine)
        self.session = db.Session(bind=self.engine)
        self.session.add(db.MetaData(name="helics_version", value=h.helicsGetVersion()))
        self.session.add(db.MetaData(name="created", value=datetime.now().isoformat()))
        self.session.add(db.SystemInfo(data=h.helicsGetSystemInfo()))

    def _setup_federate(self):
        self.name = "__observer__"
        self.fedinfo = h.helicsCreateFederateInfo()
        self.fedinfo.core_type = self._core_type
        self.fedinfo.property[h.HELICS_PROPERTY_TIME_DELTA] = 1.0
        self.fedinfo.property[h.HELICS_PROPERTY_TIME_PERIOD] = 1.0
        self.fedinfo.flag[h.HELICS_FLAG_TERMINATE_ON_ERROR] = True
        self.fedinfo.flag[h.HELICS_HANDLE_OPTION_STRICT_TYPE_CHECKING] = True
        self.fedinfo.flag[h.HELICS_FLAG_OBSERVER] = True
        logger.info("Creating observer")
        self.federate = h.helicsCreateCombinationFederate(self.name, self.fedinfo)
        logger.info("Entering initialization mode")
        self.federate.enter_initializing_mode()
        logger.info("Waiting for other federates")
        self.wait()
        self.get_data()
        self.hook()

    @property
    def federates(self) -> List[str]:
        return [cast(str, name) for name in self.federate.query("root", "federates")]

    @property
    def brokers(self) -> List[str]:
        return [cast(str, name) for name in self.federate.query("root", "brokers")]

    @property
    def publications(self) -> List[str]:
        return [cast(str, name) for name in self.federate.query("root", "publications")]

    @property
    def subscriptions(self) -> List[str]:
        return [cast(str, name) for name in self.federate.query("root", "subscriptions")]

    @property
    def inputs(self) -> List[str]:
        return [cast(str, name) for name in self.federate.query("root", "inputs")]

    def get_data(self):
        data = self.federate.query("root", "federate_map")
        data = cast(Dict, data)
        logger.info("Adding federategraph")
        self.session.add(db.FederateGraph(data=data))

        data = self.federate.query("root", "data_flow_graph")
        data = cast(Dict, data)
        logger.info("Adding datagraph")
        self.session.add(db.DataGraph(data=data))

        for core in data["cores"]:
            logger.info(f"Adding core {core}")
            self.session.add(
                db.Cores(
                    id=core["attributes"]["id"],
                    name=core["attributes"]["name"],
                    parent=core["attributes"]["parent"],
                    address=self.federate.query(core["attributes"]["name"], "address"),
                )
            )
            for federate in core["federates"]:
                assert federate["attributes"]["parent"] == core["attributes"]["id"]
                logger.info(f"Adding federate {federate}")
                self.session.add(db.Federates(id=federate["attributes"]["id"], name=federate["attributes"]["name"], parent=core["attributes"]["id"]))

                if "inputs" in federate.keys():
                    for input in federate["inputs"]:
                        logger.info(f"Adding input {input}")
                        if "sources" not in input.keys():
                            self.session.add(db.Inputs(source=input["federate"]))
                        else:
                            for s in input["sources"]:
                                self.session.add(db.Inputs(source=input["federate"], target=s["federate"]))

                if "publications" in federate.keys():
                    for publication in federate["publications"]:
                        logger.info(f"Adding publication {publication}")
                        if "targets" not in publication.keys():
                            self.session.add(db.Publications(source=publication["federate"]))
                        else:
                            for t in publication["targets"]:
                                self.session.add(db.Publications(name=publication["key"], target=publication["federate"], source=t["federate"]))

        self.session.commit()

        # Create a data table
        columns = [
            db.Column(col_name, col_type)
            for (col_name, col_type) in zip(
                self.federate.query("root", "publications"),
                [db.Float for _ in self.federate.query("root", "publications")],
            )
        ]
        columns.insert(0, db.Column("simulation_time", db.Float))
        columns.insert(0, db.Column("updated_at", db.Float))
        columns.insert(0, db.Column("id", db.Integer, db.Sequence("id"), primary_key=True, nullable=False))
        datatable = db.Table("datatable", db.Base.metadata, *columns)

        class DataTable(db.Base):
            __table__ = datatable

        self.DataTable = DataTable
        db.Base.metadata.create_all(self.engine)

    def hook(self):
        for pub in self.publications:
            s = self.federate.register_subscription(pub)
            s.set_default(0.0)

    def wait(self):
        federates = self.federates
        while not all(self.federate.query(name, "isinit") is True for name in federates):
            for name in federates:
                logger.debug(f"{name} isinit = {self.federate.query(name, 'isinit')}")
            time.sleep(1)

    def run(self):
        self.federate.enter_executing_mode()
        while True:
            for name in self.federates:
                if self.federate.query(name, "state") == "disconnected":
                    self.session.add(
                        db.FederateEventLogs(
                            name=name,
                            updated_at=time.time(),
                            state="disconnected",
                        )
                    )
                    continue
                data = cast(dict, self.federate.query(name, "current_time"))
                self.session.add(
                    db.FederateEventLogs(
                        name=name,
                        updated_at=time.time(),
                        granted_time=data["granted_time"],
                        requested_time=data["requested_time"],
                        state="executing",
                    )
                )
            simulation_time = self.federate.request_next_step()
            d = {"simulation_time": simulation_time}
            logger.info(self.federate.subscriptions)
            for k, sub in self.federate.subscriptions.items():
                d[k] = sub.value
            self.session.add(self.DataTable(updated_at=time.time(), **d))
            self.session.commit()

            if (
                all(self.federate.query(name, "state") == "disconnected" for name in self.federates if name != "__observer__")
                or simulation_time >= 9223372036.3
            ):
                break


if __name__ == "__main__":
    o = HelicsObserverFederate()
    o.run()



# -*- coding: utf-8 -*-
import os
import sys

CURRENT_DIRECTORY = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(CURRENT_DIRECTORY)
sys.path.append(os.path.dirname(CURRENT_DIRECTORY))

import pytest as pt
import helics as h
import logging


def test_python_api0():
    broker = h.helicsCreateBroker("zmq", "", "-f 1 --name=mainbroker")
    fedinfo = h.helicsCreateFederateInfo()
    assert "HelicsFederateInfo()" in repr(fedinfo)
    fedinfo.core_name = "TestFederate"
    fedinfo.core_type = "zmq"
    fedinfo.core_init = "-f 1 --broker=mainbroker"
    mFed = h.helicsCreateCombinationFederate("TestFederate", fedinfo)

    assert (
        """HelicsCombinationFederate(name = "TestFederate", state = HelicsFederateState.STARTUP, current_time = -9223372036.854776, n_publications = 0, n_subscriptions = 0, n_endpoints = 0, n_filters = 0, n_pending_messages = 0)"""
        in repr(mFed)
    )

    _ = mFed.register_endpoint("ep1")
    _ = mFed.register_global_endpoint("ep2")

    pub = mFed.register_publication("publication", h.HELICS_DATA_TYPE_STRING, "custom-units")
    assert """HelicsPublication(name = "TestFederate/publication", type = "string", units = "custom-units", info = "")""" in repr(pub)

    sub = mFed.register_subscription("TestFederate/publication", "custom-units")
    assert (
        """HelicsInput(name = "_input_3", units = "custom-units", injection_units = "", publication_type = "", type = "", target = "TestFederate/publication", info = "")"""
        in repr(sub)
    )
    assert (
        """{ 'CONNECTION_REQUIRED' = 0, 'CONNECTION_OPTIONAL' = 0, 'SINGLE_CONNECTION_ONLY' = 0, 'MULTIPLE_CONNECTIONS_ALLOWED' = 1, 'BUFFER_DATA' = 0, 'STRICT_TYPE_CHECKING' = 0, 'IGNORE_UNIT_MISMATCH' = 0, 'ONLY_TRANSMIT_ON_CHANGE' = 0, 'ONLY_UPDATE_ON_CHANGE' = 0, 'IGNORE_INTERRUPTS' = 0, 'MULTI_INPUT_HANDLING_METHOD' = 0, 'INPUT_PRIORITY_LOCATION' = -1, 'CLEAR_PRIORITY_LIST' = 1, 'CONNECTIONS' = 0 }"""
        in repr(sub.option)
    )
    sub.option["CONNECTION_REQUIRED"] = 1
    sub.option[h.HELICS_HANDLE_OPTION_CONNECTION_REQUIRED] = 1
    assert sub.option["CONNECTION_REQUIRED"] == 1

    mFed.property[h.HELICS_PROPERTY_TIME_DELTA] = 1.0
    assert mFed.property[h.HELICS_PROPERTY_TIME_DELTA] == 1.0

    mFed.property["TIME_DELTA"] = 1.0
    assert mFed.property["TIME_DELTA"] == 1.0

    mFed.enter_executing_mode()

    h.helicsCloseLibrary()

    del mFed
    del broker


def test_python_api1():

    broker = h.helicsCreateBroker("zmq", "", "-f 1 --name=mainbroker")
    fedinfo = h.helicsCreateFederateInfo()
    assert "HelicsFederateInfo()" in repr(fedinfo)
    fedinfo.core_name = "TestFederate"
    fedinfo.core_type = "zmq"
    fedinfo.core_init = "-f 1 --broker=mainbroker"
    mFed = h.helicsCreateCombinationFederate("TestFederate", fedinfo)

    assert (
        """HelicsCombinationFederate(name = "TestFederate", state = HelicsFederateState.STARTUP, current_time = -9223372036.854776, n_publications = 0, n_subscriptions = 0, n_endpoints = 0, n_filters = 0, n_pending_messages = 0)"""
        in repr(mFed)
    )

    _ = mFed.register_endpoint("ep1")
    _ = mFed.register_global_endpoint("ep2")

    pub = mFed.register_publication("publication", h.HELICS_DATA_TYPE_STRING, "custom-units")
    assert """HelicsPublication(name = "TestFederate/publication", type = "string", units = "custom-units", info = "")""" in repr(pub)

    sub = mFed.register_subscription("TestFederate/publication", "custom-units")
    assert (
        """HelicsInput(name = "_input_3", units = "custom-units", injection_units = "", publication_type = "", type = "", target = "TestFederate/publication", info = "")"""
        in repr(sub)
    )
    assert (
        """{ 'CONNECTION_REQUIRED' = 0, 'CONNECTION_OPTIONAL' = 0, 'SINGLE_CONNECTION_ONLY' = 0, 'MULTIPLE_CONNECTIONS_ALLOWED' = 1, 'BUFFER_DATA' = 0, 'STRICT_TYPE_CHECKING' = 0, 'IGNORE_UNIT_MISMATCH' = 0, 'ONLY_TRANSMIT_ON_CHANGE' = 0, 'ONLY_UPDATE_ON_CHANGE' = 0, 'IGNORE_INTERRUPTS' = 0, 'MULTI_INPUT_HANDLING_METHOD' = 0, 'INPUT_PRIORITY_LOCATION' = -1, 'CLEAR_PRIORITY_LIST' = 1, 'CONNECTIONS' = 0 }"""
        in repr(sub.option)
    )
    sub.option["CONNECTION_REQUIRED"] = 1
    assert sub.option["CONNECTION_REQUIRED"] == 1

    mFed.property[h.HELICS_PROPERTY_TIME_DELTA] = 1.0
    assert mFed.property[h.HELICS_PROPERTY_TIME_DELTA] == 1.0

    sub.set_default(b"hello")
    assert sub.bytes == b"hello"

    sub.set_default("world")
    assert sub.string == "world"

    sub.set_default(0)
    assert sub.integer == 0

    sub.set_default(True)
    assert sub.boolean is True

    sub.set_default(1.1)
    assert sub.double == 1.1

    sub.set_default(2 + 1.1j)
    assert sub.complex == 2 + 1.1j

    sub.set_default([1.0, 2.0, 3.0])
    assert sub.vector == [1.0, 2.0, 3.0]

    sub.set_default([complex(1.0, 2.0), complex(3.0, 4.0), complex(5.0, 6.0)])
    sub.info = "hello world"
    assert sub.info == "hello world"

    assert (
        """{ 'CONNECTION_REQUIRED' = 0, 'CONNECTION_OPTIONAL' = 0, 'SINGLE_CONNECTION_ONLY' = 0, 'MULTIPLE_CONNECTIONS_ALLOWED' = 1, 'BUFFER_DATA' = 0, 'STRICT_TYPE_CHECKING' = 0, 'IGNORE_UNIT_MISMATCH' = 0, 'ONLY_TRANSMIT_ON_CHANGE' = 0, 'ONLY_UPDATE_ON_CHANGE' = 0, 'IGNORE_INTERRUPTS' = 0, 'MULTI_INPUT_HANDLING_METHOD' = 0, 'INPUT_PRIORITY_LOCATION' = 0, 'CLEAR_PRIORITY_LIST' = 0, 'CONNECTIONS' = 0 }"""
        in repr(mFed.publications["TestFederate/publication"].option)
    )
    mFed.publications["TestFederate/publication"].option["CONNECTION_REQUIRED"] = 1
    assert mFed.publications["TestFederate/publication"].option["CONNECTION_REQUIRED"] == 1

    mFed.enter_executing_mode()

    data = "random-data"

    mFed.endpoints["TestFederate/ep1"].default_destination = "ep2"
    mFed.endpoints["TestFederate/ep1"].info = "information"

    assert mFed.endpoints["TestFederate/ep1"].default_destination == "ep2"
    assert mFed.endpoints["TestFederate/ep1"].info == "information"

    mFed.endpoints["TestFederate/ep1"].send_data(data, "ep2", 1.0)

    mFed.publications["TestFederate/publication"].publish("first-time")

    assert mFed.request_time(2.0) == 1.0

    assert mFed.subscriptions["TestFederate/publication"].value == "first-time"
    try:
        assert mFed.subscriptions["TestFederate/publication"].bytes == b"first-time"
    except Exception:
        # TODO: this does not work as expected
        with pt.raises(AssertionError):
            assert mFed.subscriptions["TestFederate/publication"].bytes == b"first-time"

    assert mFed.has_message()

    assert mFed.endpoints["TestFederate/ep1"].has_message() is False

    assert mFed.endpoints["ep2"].has_message()

    message = mFed.endpoints["ep2"].get_message()

    assert message.message_id == 55
    assert message.is_valid() is True
    assert message.data == "random-data"
    assert message.raw_data == b"random-data"
    assert len(message.raw_data) == 11
    assert message.original_destination == ""
    assert message.original_source == "TestFederate/ep1"
    assert message.source == "TestFederate/ep1"
    assert message.time == 1.0

    assert (
        """HelicsMessage(source = "TestFederate/ep1", destination = "ep2", original_source = "TestFederate/ep1", original_destination = "", time = 1.0, id = 55, message = "random-data")"""
        in repr(message)
    )

    message.append("-random")
    assert message.is_valid() is True
    assert message.data == "random-data-random"

    message.data = "random-data"
    assert message.raw_data == b"random-data"

    message.raw_data = b"random-data-random"
    assert message.data == "random-data-random"

    assert message.is_valid() is True
    message.message_id = 100
    message.source = "earth"
    message.destination = "moon"
    message.original_source = "hello-world"
    message.original_destination = "goodbye-world"
    message.time = 2.0

    assert message.is_valid() is True
    assert message.message_id == 100
    assert message.source == "earth"
    assert message.destination == "moon"
    assert message.original_source == "hello-world"
    assert message.original_destination == "goodbye-world"
    assert message.time == 2.0

    assert message.is_valid() is True
    assert (
        """<{ 1 = False, 2 = False, 3 = False, 4 = False, 5 = False, 6 = False, 7 = False, 8 = False, 9 = False, 10 = False, 11 = False, 12 = False, 13 = False, 14 = False, 15 = False }>"""
        in repr(message.flag)
    )
    assert message.is_valid() is True
    message.flag[1] = True
    assert message.flag[1] is True

    mFed.publications["TestFederate/publication"].publish(1)

    assert mFed.request_next_step() == 2.0

    assert mFed.subscriptions["TestFederate/publication"].string == "1"

    mFed.endpoints["TestFederate/ep1"].send_data(message, "ep2")
    mFed.endpoints["TestFederate/ep1"].send_data(message, "ep2", 1.0)

    mFed.publications["TestFederate/publication"].publish(1 + 2j)

    assert mFed.request_next_step() == 3.0

    try:
        assert mFed.subscriptions["TestFederate/publication"].string == "1+2j"
    except AssertionError:
        assert mFed.subscriptions["TestFederate/publication"].string == "[1,2]"

    mFed.publications["TestFederate/publication"].publish([1, 2, 3, 4, 5])

    assert mFed.request_next_step() == 4.0

    assert mFed.subscriptions["TestFederate/publication"].vector == [
        1.0,
        2.0,
        3.0,
        4.0,
        5.0,
    ]

    mFed.publications["TestFederate/publication"].publish(False)

    assert mFed.request_next_step() == 5.0

    assert mFed.subscriptions["TestFederate/publication"].boolean is False
    m = mFed.create_message()
    assert (
        """HelicsMessage(source = "", destination = "", original_source = "", original_destination = "", time = 0.0, id = 0, message = "")"""
        in repr(m)
    )
    mFed.info = "hello-world"
    assert mFed.info == "hello-world"

    m = mFed.endpoints["ep2"].create_message()
    assert (
        """HelicsMessage(source = "", destination = "", original_source = "", original_destination = "", time = 0.0, id = 0, message = "")"""
        in repr(m)
    )

    mFed.disconnect()

    del mFed
    del broker


def test_python_api2():

    broker = h.helicsCreateBroker("zmq", "broker", "--federates 1 --loglevel=warning")
    assert broker.is_connected()

    broker.set_global("hello", "world")

    broker.data_link("hello", "world")
    broker.add_destination_filter_to_endpoint("hello", "world")
    broker.add_source_filter_to_endpoint("hello", "world")
    try:
        assert broker.query("hello", "world") == "#invalid"
    except AssertionError:
        assert broker.query("hello", "world") == {"error": {"code": 404, "message": "query not valid"}}

    fi = h.helicsCreateFederateInfo()
    fi.core_init = "--federates 1"
    fi.property[h.HELICS_PROPERTY_INT_LOG_LEVEL] = 2

    fed = h.helicsCreateCombinationFederate("test1", fi)

    assert "HelicsCore" in repr(fed.core)
    assert 'address = "tcp://127.0.0.1' in repr(fed.core)

    assert fed.core.is_connected()
    fed.core.set_ready_to_init()

    assert "n_publications = 0" in repr(fed)
    assert "n_subscriptions = 0" in repr(fed)
    assert "n_endpoints = 0" in repr(fed)
    assert "n_filters = 0" in repr(fed)

    assert fed.property["DELTA"] == 1e-09
    assert fed.property["PERIOD"] == 0.0
    assert fed.property["OFFSET"] == 0.0
    assert fed.property["RT_LAG"] == 0.0
    assert fed.property["RT_LEAD"] == 0.0
    assert fed.property["RT_TOLERANCE"] == 0.0
    assert fed.property["INPUT_DELAY"] == 0.0
    assert fed.property["OUTPUT_DELAY"] == 0.0
    assert fed.property["MAX_ITERATIONS"] == 50
    assert fed.property["LOG_LEVEL"] == 2
    assert fed.property["FILE_LOG_LEVEL"] == 2
    assert fed.property["CONSOLE_LOG_LEVEL"] == 2

    fed.property[h.HELICS_PROPERTY_INT_LOG_LEVEL] = 5

    fed.separator = "_"
    assert fed.separator == "_"

    fed.separator = "/"
    assert fed.separator == "/"

    assert fed.property[h.HelicsProperty.TIME_DELTA] == 1e-09
    assert fed.property[h.HelicsProperty.TIME_PERIOD] == 0.0
    assert fed.property[h.HelicsProperty.TIME_OFFSET] == 0.0
    assert fed.property[h.HelicsProperty.TIME_RT_LAG] == 0.0
    assert fed.property[h.HelicsProperty.TIME_RT_LEAD] == 0.0
    assert fed.property[h.HelicsProperty.TIME_RT_TOLERANCE] == 0.0
    assert fed.property[h.HelicsProperty.TIME_INPUT_DELAY] == 0.0
    assert fed.property[h.HelicsProperty.TIME_OUTPUT_DELAY] == 0.0
    assert fed.property[h.HelicsProperty.INT_MAX_ITERATIONS] == 50
    assert fed.property[h.HelicsProperty.INT_LOG_LEVEL] == 5
    assert fed.property[h.HelicsProperty.INT_FILE_LOG_LEVEL] == 5
    assert fed.property[h.HelicsProperty.INT_CONSOLE_LOG_LEVEL] == 5

    assert fed.property[h.HelicsProperty.TIME_DELTA.value] == 1e-09
    assert fed.property[h.HelicsProperty.TIME_PERIOD.value] == 0.0
    assert fed.property[h.HelicsProperty.TIME_OFFSET.value] == 0.0
    assert fed.property[h.HelicsProperty.TIME_RT_LAG.value] == 0.0
    assert fed.property[h.HelicsProperty.TIME_RT_LEAD.value] == 0.0
    assert fed.property[h.HelicsProperty.TIME_RT_TOLERANCE.value] == 0.0
    assert fed.property[h.HelicsProperty.TIME_INPUT_DELAY.value] == 0.0
    assert fed.property[h.HelicsProperty.TIME_OUTPUT_DELAY.value] == 0.0
    assert fed.property[h.HelicsProperty.INT_MAX_ITERATIONS.value] == 50
    assert fed.property[h.HelicsProperty.INT_LOG_LEVEL.value] == 5
    assert fed.property[h.HelicsProperty.INT_FILE_LOG_LEVEL.value] == 5
    assert fed.property[h.HelicsProperty.INT_CONSOLE_LOG_LEVEL.value] == 5

    assert "'TIME_DELTA' = 1e-09" in repr(fed.property)
    assert "'TIME_PERIOD' = 0.0" in repr(fed.property)
    assert "'TIME_OFFSET' = 0.0" in repr(fed.property)
    assert "'TIME_RT_LAG' = 0.0" in repr(fed.property)
    assert "'TIME_RT_LEAD' = 0.0" in repr(fed.property)
    assert "'TIME_RT_TOLERANCE' = 0.0" in repr(fed.property)
    assert "'TIME_INPUT_DELAY' = 0.0" in repr(fed.property)
    assert "'TIME_OUTPUT_DELAY' = 0.0" in repr(fed.property)
    assert "'INT_MAX_ITERATIONS' = 50" in repr(fed.property)
    assert "'INT_LOG_LEVEL' = 5" in repr(fed.property)
    assert "'INT_FILE_LOG_LEVEL' = 5" in repr(fed.property)
    assert "'INT_CONSOLE_LOG_LEVEL' = 5" in repr(fed.property)

    assert fed.flag[h.HELICS_FLAG_OBSERVER] is False
    assert fed.flag[h.HELICS_FLAG_UNINTERRUPTIBLE] is False
    assert fed.flag[h.HELICS_FLAG_INTERRUPTIBLE] is True
    assert fed.flag[h.HELICS_FLAG_SOURCE_ONLY] is False
    assert fed.flag[h.HELICS_FLAG_ONLY_TRANSMIT_ON_CHANGE] is False
    assert fed.flag[h.HELICS_FLAG_ONLY_UPDATE_ON_CHANGE] is False
    assert fed.flag[h.HELICS_FLAG_WAIT_FOR_CURRENT_TIME_UPDATE] is False
    assert fed.flag[h.HELICS_FLAG_RESTRICTIVE_TIME_POLICY] is False
    assert fed.flag[h.HELICS_FLAG_REALTIME] is False
    assert fed.flag[h.HELICS_FLAG_SLOW_RESPONDING] is False
    assert fed.flag[h.HELICS_FLAG_IGNORE_TIME_MISMATCH_WARNINGS] is False
    assert fed.flag[h.HELICS_FLAG_TERMINATE_ON_ERROR] is False

    assert fed.flag["TERMINATE_ON_ERROR"] is False
    fed.flag["TERMINATE_ON_ERROR"] = False

    assert fed.flag[h.HELICS_FLAG_OBSERVER.value] is False
    assert fed.flag[h.HELICS_FLAG_UNINTERRUPTIBLE.value] is False
    assert fed.flag[h.HELICS_FLAG_INTERRUPTIBLE.value] is True
    assert fed.flag[h.HELICS_FLAG_SOURCE_ONLY.value] is False
    assert fed.flag[h.HELICS_FLAG_ONLY_TRANSMIT_ON_CHANGE.value] is False
    assert fed.flag[h.HELICS_FLAG_ONLY_UPDATE_ON_CHANGE.value] is False
    assert fed.flag[h.HELICS_FLAG_WAIT_FOR_CURRENT_TIME_UPDATE.value] is False
    assert fed.flag[h.HELICS_FLAG_RESTRICTIVE_TIME_POLICY.value] is False
    assert fed.flag[h.HELICS_FLAG_REALTIME.value] is False
    assert fed.flag[h.HELICS_FLAG_SLOW_RESPONDING.value] is False
    assert fed.flag[h.HELICS_FLAG_IGNORE_TIME_MISMATCH_WARNINGS.value] is False
    assert fed.flag[h.HELICS_FLAG_TERMINATE_ON_ERROR.value] is False

    assert "'OBSERVER' = False" in repr(fed.flag)
    assert "'UNINTERRUPTIBLE' = False" in repr(fed.flag)
    assert "'INTERRUPTIBLE' = True" in repr(fed.flag)
    assert "'SOURCE_ONLY' = False" in repr(fed.flag)
    assert "'ONLY_TRANSMIT_ON_CHANGE' = False" in repr(fed.flag)
    assert "'ONLY_UPDATE_ON_CHANGE' = False" in repr(fed.flag)
    assert "'WAIT_FOR_CURRENT_TIME_UPDATE' = False" in repr(fed.flag)
    assert "'RESTRICTIVE_TIME_POLICY' = False" in repr(fed.flag)
    assert "'REALTIME' = False" in repr(fed.flag)
    assert "'SLOW_RESPONDING' = False," in repr(fed.flag)
    assert "'IGNORE_TIME_MISMATCH_WARNINGS' = False," in repr(fed.flag)
    assert "'TERMINATE_ON_ERROR' = False" in repr(fed.flag)

    fed.flag[h.HELICS_FLAG_TERMINATE_ON_ERROR] = True

    assert fed.flag[h.HELICS_FLAG_TERMINATE_ON_ERROR] is True

    fed.core.register_filter(h.HelicsFilterType.DELAY, "")
    fed.core.register_cloning_filter("")

    fed.register_filter(h.HelicsFilterType.DELAY, "")
    fed.register_cloning_filter("")

    fed.register_global_filter(h.HelicsFilterType.DELAY, "")
    fed.register_global_cloning_filter("")

    fed.core.set_global("hello", "world")

    try:
        assert fed.core.query("broker", "something") == "#invalid"
    except AssertionError:
        assert fed.core.query("broker", "something") == {"error": {"code": 400, "message": "unrecognized broker query"}}

    fed.add_dependency("hello")

    fed.core.disconnect()

    assert fed.core.wait_for_disconnect()

    del fed

    broker.disconnect()
    assert broker.wait_for_disconnect()

    del broker

    h.helicsCleanupLibrary()
    h.helicsCloseLibrary()


def test_python_api3():
    core1 = h.helicsCreateCore("inproc", "core1", "--autobroker")

    assert """HelicsCore(identifier = "core1", address = "core1")""" in repr(core1)

    core2 = core1.clone()

    assert core1.identifier == "core1"

    source_filter1 = core1.register_filter(h.HELICS_FILTER_TYPE_DELAY, "core1SourceFilter")

    source_filter1.add_source_target("ep1")

    assert (
        """<{ 'CONNECTION_REQUIRED' = 0, 'CONNECTION_OPTIONAL' = 0, 'SINGLE_CONNECTION_ONLY' = 0, 'MULTIPLE_CONNECTIONS_ALLOWED' = 0, 'BUFFER_DATA' = 0, 'STRICT_TYPE_CHECKING' = 0, 'IGNORE_UNIT_MISMATCH' = 0, 'ONLY_TRANSMIT_ON_CHANGE' = 0, 'ONLY_UPDATE_ON_CHANGE' = 0, 'IGNORE_INTERRUPTS' = 0, 'MULTI_INPUT_HANDLING_METHOD' = 0, 'INPUT_PRIORITY_LOCATION' = 0, 'CLEAR_PRIORITY_LIST' = 0, 'CONNECTIONS' = 0 }>"""
        in repr(source_filter1.option)
    )

    source_filter1.option["CONNECTION_REQUIRED"] = 1
    assert source_filter1.option["CONNECTION_REQUIRED"] == 1

    source_filter1.add_source_target("hello")
    source_filter1.add_destination_target("world")
    source_filter1.remove_destination_target("world")

    source_filter1.info = "hello world"
    assert source_filter1.info == "hello world"

    source_filter1.set("hello", 1)

    destination_filter1 = core1.register_filter(h.HELICS_FILTER_TYPE_DELAY, "core1DestinationFilter")

    destination_filter1.add_destination_target("ep2")
    cloning_filter1 = core1.register_cloning_filter("ep3")

    cloning_filter1.remove_delivery_endpoint("ep3")
    cloning_filter1.add_delivery_endpoint("ep3")

    assert core1.is_valid()
    assert core1.is_connected()
    core1.set_ready_to_init()

    core1.disconnect()
    core2.disconnect()

    del core1
    del core2

    h.helicsCloseLibrary()


def test_python_api4():
    fi = h.helicsCreateFederateInfo()
    fi.separator = "_"
    fi.broker = "broker test"

    fi.broker_key = "hello-world"

    fi.broker_port = 8929
    fi.local_port = 8229

    fi.broker_init = "-f 3"

    fi.flag[h.HelicsFlag.TERMINATE_ON_ERROR] = True
    fi.property[h.HelicsProperty.TIME_DELTA] = 1.0
    fi.property["TIME_DELTA"] = 1.0


def test_python_api5():
    broker = h.helicsCreateBroker("zmq", "broker", "--federates 1")
    fi = h.helicsCreateFederateInfo()

    fed = h.helicsCreateCombinationFederate("test1", fi)
    with pt.raises(h.HelicsException):
        fed.register_interfaces("unknownfile.json")

    fed.core.disconnect()

    assert fed.core.wait_for_disconnect()

    del fed

    broker.disconnect()
    assert broker.wait_for_disconnect()

    del broker

    h.helicsCleanupLibrary()
    h.helicsCloseLibrary()


def test_python_api6():
    broker = h.helicsCreateBroker("zmq", "broker", "--federates 1")
    fi = h.helicsCreateFederateInfo()

    fed = h.helicsCreateCombinationFederate("test1", fi)

    fed.enter_initializing_mode()
    fed.enter_executing_mode()

    fed.core.disconnect()
    assert fed.core.wait_for_disconnect()
    del fed

    broker.disconnect()
    assert broker.wait_for_disconnect()
    del broker

    h.helicsCleanupLibrary()
    h.helicsCloseLibrary()


@pt.mark.skip(reason="Fails to pass on windows and linux")
def test_python_api7():
    broker = h.helicsCreateBroker("zmq", "broker", "--federates 1")
    fi = h.helicsCreateFederateInfo()

    fed = h.helicsCreateCombinationFederate("test1", fi)

    _ = fed.register_filter(h.HELICS_FILTER_TYPE_DELAY, "core1SourceFilter")

    assert fed.get_filter_by_name("core1SourceFilter").name == fed.get_filter_by_index(0).name

    fed.set_global("hello", "world")

    # TODO: is_async_operation_completed
    assert fed.is_async_operation_completed() is False

    fed.enter_initializing_mode_async()
    fed.enter_initializing_mode_complete()

    assert fed.is_async_operation_completed() is False

    fed.enter_executing_mode_async()
    fed.enter_executing_mode_complete()

    assert fed.is_async_operation_completed() is False

    fed.request_time_advance(2.0)
    assert fed.current_time == 2.0

    fed.request_time_async(4.0)
    assert fed.current_time == 2.0
    fed.request_time_complete()
    assert fed.current_time == 4.0

    try:
        assert fed.query("hello", "world") == "#disconnected"
    except AssertionError:
        assert fed.query("hello", "world") == {"error": {"code": 404, "message": "query not valid"}}

    fed.local_error(0, "local")
    fed.global_error(0, "global")

    fed.log_message("error", logging.ERROR)
    fed.log_message("warn", logging.WARN)
    fed.log_message("info", logging.INFO)
    fed.log_message("debug", logging.DEBUG)
    fed.log_message("summary", h.HELICS_LOG_LEVEL_SUMMARY)

    fed.disconnect_async()
    fed.disconnect_complete()

    fed.core.disconnect()
    assert fed.core.wait_for_disconnect()
    del fed

    broker.disconnect()
    assert broker.wait_for_disconnect()
    del broker

    h.helicsCleanupLibrary()
    h.helicsCloseLibrary()


def test_python_api8():

    broker = h.helicsCreateBroker("zmq", "", "-f 1 --name=mainbroker")

    cfed = h.helicsCreateCombinationFederateFromConfig(os.path.join(CURRENT_DIRECTORY, "combinationfederate.json"))

    assert len(cfed.endpoints) == 2
    assert len(cfed.subscriptions) == 2

    h.helicsFederateDestroy(cfed)
    h.helicsFederateFree(cfed)
    h.helicsBrokerDestroy(broker)
    h.helicsCloseLibrary()


def test_python_api9():

    broker = h.helicsCreateBroker("zmq", "", "-f 1 --name=mainbroker")
    fedinfo = h.helicsCreateFederateInfo()
    assert "HelicsFederateInfo()" in repr(fedinfo)
    fedinfo.core_name = "TestFederate"
    fedinfo.core_type = "zmq"
    fedinfo.core_init = "-f 1 --broker=mainbroker"
    mFed = h.helicsCreateCombinationFederate("TestFederate", fedinfo)

    pub = mFed.register_publication("publication", h.HELICS_DATA_TYPE_COMPLEX_VECTOR, "custom-units")
    assert pub.type == "complex_vector"

    sub = mFed.register_subscription("TestFederate/publication", "custom-units")
    sub.set_default([complex(1.0, 2.0), complex(3.0, 4.0), complex(5.0, 6.0)])

    mFed.enter_executing_mode()

    assert sub.publication_type == "complex_vector"
    print(sub.value)

    mFed.disconnect()

    del mFed
    del broker



import os
import subprocess
import sys

from ._build import PYHELICS_INSTALL

BIN_DIR = os.path.join(PYHELICS_INSTALL, 'bin')

def _program(name, args):
    return subprocess.call([os.path.join(BIN_DIR, name)] + args)

def helics_app():
    raise SystemExit(_program('helics_app', sys.argv[1:]))

def helics_broker():
    raise SystemExit(_program('helics_broker', sys.argv[1:]))

def helics_broker_server():
    raise SystemExit(_program('helics_broker_server', sys.argv[1:]))

def helics_player():
    raise SystemExit(_program('helics_player', sys.argv[1:]))

def helics_recorder():
    raise SystemExit(_program('helics_recorder', sys.argv[1:]))



# -*- coding: utf-8 -*-
import click


def info(msg, blink=False):
    echo(msg, fg="green", level="info", blink=blink)


def warn(msg, blink=True):
    echo(msg, blink=blink)


def error(msg, blink=False):
    echo(msg, fg="red", level="error", blink=blink)


def echo(msg, fg="yellow", level="warn", blink=True):
    click.echo(click.style("[", fg=fg) + click.style(level, fg=fg, blink=blink) + click.style("] ", fg=fg) + click.style(msg, fg=fg))



# -*- coding: utf-8 -*-
import time

from math import pi
import helics as h
import random


def main(n):

    federate_name = f"SenderFederate{n}"

    print(f"{federate_name}: Helics version = {h.helicsGetVersion()}")

    fedinfo = h.helicsCreateFederateInfo()
    fedinfo.core_type = "zmq"
    fedinfo.property[h.HELICS_PROPERTY_TIME_DELTA] = 1.0
    fedinfo.property[h.HELICS_PROPERTY_TIME_PERIOD] = 1.0

    fed = h.helicsCreateCombinationFederate(federate_name, fedinfo)

    print(f"{federate_name}: Combination federate created")

    fed.register_global_publication(f"globaltopic{n}", h.HELICS_DATA_TYPE_DOUBLE)

    print(f"{federate_name}: Publication registered")

    fed.enter_executing_mode()

    print(f"{federate_name}: Entering execution mode")

    for t in range(0, 10):
        value = pi * t

        fed.request_next_step()

        fed.publications[f"globaltopic{n}"].publish(value)

        print(f"{federate_name}: Sending value pi = {value} at time {fed.current_time}")

        # Computing user needs
        time.sleep(float(n) * (1 + (0.5 - random.random()) / 10))

    fed.disconnect()
    print(f"{federate_name}: Federate finalized")


if __name__ == "__main__":
    import sys

    if len(sys.argv) == 1:
        n = 1
    else:
        n = int(sys.argv[1])
    main(n)



# -*- coding: utf-8 -*-
import helics as h


def main(n):
    federate_name = f"ReceiverFederate{n}"

    print(f"{federate_name}: Helics version = {h.helicsGetVersion()}")

    fedinfo = h.helicsCreateFederateInfo()
    fedinfo.core_type = "zmq"
    fedinfo.property[h.HELICS_PROPERTY_TIME_DELTA] = 1.0
    fedinfo.property[h.HELICS_PROPERTY_TIME_PERIOD] = 1.0

    fed = h.helicsCreateCombinationFederate(federate_name, fedinfo)

    print(f"{federate_name}: Value federate created")

    fed.register_subscription(f"globaltopic{n}").set_default(0.0)

    fed.enter_executing_mode()
    print(f"{federate_name}: Entering execution mode")

    currenttime = -1
    while currenttime <= 10:
        if currenttime < 10:
            currenttime = fed.request_time(10)
        else:
            break
        value = fed.subscriptions[f"globaltopic{n}"].value
        print(f"{federate_name}: Received value = {value} at time {currenttime}")

    fed.disconnect()

    print(f"{federate_name}: Federate finalized")


if __name__ == "__main__":
    import sys

    if len(sys.argv) == 1:
        n = 1
    else:
        n = int(sys.argv[1])
    main(n)



# -*- coding: utf-8 -*-
"""
HELICS command line interface
"""

import json
import os
import io
import shlex
import subprocess
import collections
import urllib.request
import logging
from ._version import __version__
from .status_checker import CheckStatusThread, HELICSRuntimeError

import click
import pathlib

from typing import Union

from .utils import echo, info, warn, error

HELICS_CLI_SERVER_API = "http://127.0.0.1:5000/api"

logger = logging.getLogger(__name__)


def _get_version():
    try:
        import helics as h

        helics_version = "Python HELICS version {}\n\nHELICS Library version {}".format(h.__version__, h.helicsGetVersion())
    except ImportError:
        helics_version = "Python `helics` package not installed. Install using `pip install helics --upgrade`."
    try:
        import helics_apps as ha

        helics_apps_version = "HELICS Apps version {}".format(ha.__version__)

    except ImportError:
        helics_apps_version = ""

    try:
        import helics as h
        import helics_apps as ha

        if not h.helicsGetVersion().startswith(ha.__version__.strip("v")):
            echo("`helics` and `helics-apps` versions don't match. You may want to run `pip install helics helics-apps --upgrade`.")
    except ImportError:
        pass

    try:
        import flask
    except ImportError:
        echo('helics-cli\'s web interface is not installed. You may want to run `pip install "helics[cli]"`.')

    try:
        import sqlalchemy
    except ImportError:
        echo('helics-cli\'s observer functionality is not installed. You may want to run `pip install "helics[cli]"`.')

    return """{}

{}

{}
""".format(
        __version__, helics_version, helics_apps_version
    ).strip()


VERSION = _get_version()


@click.group()
@click.version_option(VERSION, "--version")
@click.option("--verbose", "-v", count=True)
@click.pass_context
def cli(ctx, verbose):
    """
    HELICS Runner command line interface
    """
    ctx.obj = {}
    ctx.obj["verbose"] = verbose


@cli.command()
@click.option(
    "--open/--no-open",
    is_flag=True,
    default=True,
    show_default=True,
    help="Open browser on startup",
)
def server(open: bool):
    import webbrowser
    from . import flaskr

    if open:
        webbrowser.open("http://127.0.0.1:5000", 1)
    flaskr.run()


@cli.command()
@click.option("--db-folder", prompt="path to database folder", type=click.Path(exists=True, file_okay=False, writable=True, path_type=pathlib.Path))
def observer(db_folder: pathlib.Path):
    from .observer import HelicsObserverFederate

    o = HelicsObserverFederate(folder=db_folder)
    o.run()


@cli.command()
@click.option(
    "--path",
    required=True,
    type=click.Path(file_okay=True, exists=True),
    help="Path to profile.txt that describes profiling results of a federation",
)
@click.option(
    "--invert",
    is_flag=True,
    default=True,
    help="Invert plot",
)
@click.option("--save", prompt=True, prompt_required=False, type=click.Path(), default=None, help="Path to save the plot")
def profile_plot(path, save, invert):
    from . import profile as p

    p.plot(p.profile(path, invert), save=save, kind="realtime")


from dataclasses import dataclass


@dataclass
class Job:
    name: str
    process: subprocess.Popen
    file: str


@dataclass
class Output:
    name: str
    file: Union[io.TextIOWrapper, None]


def post(url, data):
    r = urllib.request.Request("{}{}".format(HELICS_CLI_SERVER_API, url))
    r.add_header("Content-Type", "application/json; charset=utf-8")
    bytes = json.dumps(data).encode("utf-8")
    r.add_header("Content-Length", str(len(bytes)))
    try:
        with urllib.request.urlopen(r, bytes) as response:
            return json.loads(response.read().decode(response.info().get_param("charset") or "utf-8"))
    except Exception as e:
        logger.exception("Unable to post to helics-cli server: {}".format(e))


@cli.command()
@click.option(
    "--path",
    required=True,
    type=click.Path(file_okay=True, exists=True),
    help="Path to config.json that describes how to run a federation",
)
@click.option("--silent", is_flag=True)
@click.option("--connect-server", is_flag=True)
@click.option("--no-log-files", is_flag=True, default=False)
@click.option("--no-kill-on-error", is_flag=True, default=False, help="Do not kill all federates on error")
def run(path, silent, connect_server, no_log_files, no_kill_on_error):
    """
    Run HELICS federation
    """

    r = urllib.request.Request("{}/health".format(HELICS_CLI_SERVER_API))

    helics_server_available = False
    try:
        if connect_server:
            with urllib.request.urlopen(r) as response:
                helics_server_available = (
                    json.loads(response.read().decode(response.info().get_param("charset") or "utf-8")).get("status", None) == 200
                )
    except Exception:
        warn("Unable to connect to helics-cli web server")
        helics_server_available = False

    log = not no_log_files
    kill_on_error = not no_kill_on_error
    path_to_config = os.path.abspath(path)
    path = os.path.dirname(path_to_config)

    if not os.path.exists(path_to_config):
        info(
            "Unable to find file `config.json` in path: {path_to_config}".format(path_to_config=path_to_config),
        )
        return None

    with open(path_to_config) as f:
        config = json.loads(f.read())

    if not silent:
        info("Running federation: {name}".format(name=config["name"]))

    if "broker" in config.keys() and config["broker"] is not False:
        if not silent:
            info(
                "Adding auto broker (i.e. `helics_broker -f{f}`) to helics-cli subprocesses.".format(f=len(config["federates"])),
                blink=True,
            )
        config["federates"].append(
            {"directory": ".", "exec": "helics_broker -f{}".format(len(config["federates"])), "host": "localhost", "name": "broker"}
        )

    names = [c["name"] for c in config["federates"]]
    if len(set(n for n in names)) != len(config["federates"]):
        error("Repeated names found in runner.json federates.", blink=True)
        for n, c in [(item, count) for item, count in collections.Counter(names).items() if count > 1]:
            info('Found name "{}" {} times'.format(n, c))
        return -1

    if helics_server_available:
        post("/runner/file/name", {"name": os.path.basename(path_to_config)})
        post("/runner/file/folder", {"folder": os.path.dirname(path_to_config)})

    process_list = []
    output_list = []
    for f in config["federates"]:
        if not silent:
            info(
                "Running federate {name} as a background process".format(name=f["name"]),
            )

        fname = os.path.abspath(os.path.join(path, "{}.log".format(f["name"])))
        if log is True:
            o = Output(fname, open(fname, "w"))
        else:
            o = Output(fname, None)

        try:
            directory = os.path.join(path, f["directory"])

            env = dict(os.environ)
            if "env" in f:
                for k, v in f["env"].items():
                    env[k] = v
            p = subprocess.Popen(
                shlex.split(f["exec"]),
                cwd=os.path.abspath(os.path.expanduser(directory)),
                stdout=o.file,
                stderr=o.file,
                env=env,
            )

            p = Job(name=f["name"], process=p, file=o.name)

        except FileNotFoundError as e:
            raise click.ClickException("FileNotFoundError: {}".format(e))
        process_list.append(p)
        if o.file is not None:
            output_list.append(o)

    t = CheckStatusThread(process_list, kill_on_error, helics_server_available)

    try:
        t.start()
        if not silent:
            info(
                "Waiting for {} processes to finish ...".format(len(process_list)),
            )
        for p in process_list:
            p.process.wait()
    except KeyboardInterrupt:
        warn("User interrupted processes. Terminating safely ...")
        for o in output_list:
            o.file.close()
        for p in process_list:
            p.process.kill()
    except Exception as e:
        click.echo("")
        error(f"{e}. Terminating ...")
        if kill_on_error:
            for o in output_list:
                o.file.close()
            for p in process_list:
                p.process.kill()
    finally:
        for p in process_list:
            t.status(p)
            if p.process.returncode != 0 and p.process.returncode != -9 and p.process.returncode is not None:
                error("Process {} exited with return code {}".format(p.name, p.process.returncode))
                if os.path.exists(p.file):
                    with open(p.file) as f:
                        warn("Last 10 lines of {}.log:".format(p.name), blink=False)
                        print("...")
                        for line in f.readlines()[-10:]:
                            print(line, end="")
                        print("...")
    info("Done.")


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

import re
import sys
import os
import struct
import shutil
import platform
import tarfile
import zipfile
import subprocess
import shlex

from setuptools import setup, Extension, Command
from setuptools.dist import Distribution
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
from setuptools.command.sdist import sdist

from wheel.bdist_wheel import bdist_wheel
from distutils.dir_util import copy_tree
from distutils import log

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen


HERE = os.path.dirname(os.path.abspath(__file__))
IS_REPO = os.path.exists(os.path.join(HERE, ".git"))
STATIC_DIR = os.path.join(HERE, "helics", "static")
NODE_ROOT = os.path.join(HERE, "client")
NPM_PATH = os.pathsep.join(
    [
        os.path.join(NODE_ROOT, "node_modules", ".bin"),
        os.environ.get("PATH", os.defpath),
    ]
)


def update_package_data(distribution):
    """update package_data to catch changes during setup"""
    build_py = distribution.get_command_obj("build_py")
    # distribution.package_data = find_package_data()
    # re-init build_py options which load package_data
    build_py.finalize_options()


def js_prerelease(command, strict=False):
    """decorator for building minified js/css prior to another command"""

    class DecoratedCommand(command):
        def run(self):
            jsdeps = self.distribution.get_command_obj("jsdeps")
            if not IS_REPO and all(os.path.exists(t) for t in jsdeps.targets):
                # sdist, nothing to do
                command.run(self)
                return

            try:
                self.distribution.run_command("jsdeps")
            except Exception as e:
                missing = [t for t in jsdeps.targets if not os.path.exists(t)]
                if strict or missing:
                    log.warn("rebuilding js and css failed")
                    if missing:
                        log.error("missing files: %s" % missing)
                    raise e
                else:
                    log.warn("rebuilding js and css failed (not a problem)")
                    log.warn(str(e))
            command.run(self)
            update_package_data(self.distribution)

    return DecoratedCommand


class NPM(Command):
    description = "install package.json dependencies using npm"

    user_options = []

    node_modules = os.path.join(NODE_ROOT, "node_modules")

    targets = ["index.html"]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def get_npm_name(self):
        npm_name = "npm"
        if platform.system() == "Windows":
            npm_name = "npm.cmd"
        return npm_name

    def has_npm(self):
        npm_name = self.get_npm_name()
        try:
            subprocess.check_call([npm_name, "--version"])
            return True
        except:
            return False

    def should_run_npm_install(self):
        node_modules_exists = os.path.exists(self.node_modules)
        return self.has_npm() and not node_modules_exists

    def run(self):
        has_npm = self.has_npm()
        if not has_npm:
            log.error("`npm` unavailable.  If you're running this command using " "sudo, make sure `npm` is available to sudo")

        env = os.environ.copy()
        env["PATH"] = NPM_PATH

        npm_name = self.get_npm_name()

        if self.should_run_npm_install():
            log.info("Installing build dependencies with npm.  " "This may take a while...")
            subprocess.check_call(
                [npm_name, "install"],
                cwd=NODE_ROOT,
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
            os.utime(self.node_modules, None)

        subprocess.check_call(
            [npm_name, "run", "build"],
            cwd=NODE_ROOT,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )

        copy_tree(os.path.join(NODE_ROOT, "build"), os.path.join(STATIC_DIR))

        for t in self.targets:
            if not os.path.exists(os.path.join(STATIC_DIR, t)):
                msg = "Missing file: %s" % t
                if not has_npm:
                    msg += "\nnpm is required to build a development version "
                    "of a widget extension"
                raise ValueError(msg)

        # update package data in case this created new files
        update_package_data(self.distribution)


def read(*names, **kwargs):
    with io.open(join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")) as fh:
        return fh.read()


PYHELICS_VERSION = read(os.path.join(os.path.dirname(__file__), "helics", "_version.py"), encoding="utf-8")
PYHELICS_VERSION = PYHELICS_VERSION.splitlines()[1].split()[2].strip('"').strip("'").lstrip("v")

HELICS_VERSION = re.findall(r"(?:(\d+\.(?:\d+\.)*\d+))", PYHELICS_VERSION)[0]
# HELICS_VERSION = "{}-beta".format(HELICS_VERSION)

CURRENT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

HELICS_SOURCE = os.path.join(CURRENT_DIRECTORY, "./_source")
PYHELICS_INSTALL = os.path.join(CURRENT_DIRECTORY, "./helics/install")

DOWNLOAD_URL = "https://github.com/GMLC-TDC/HELICS/releases/download/v{version}/Helics-v{version}-source.tar.gz".format(version=HELICS_VERSION)


def create_default_url(helics_version, plat_name=""):
    if "macos" in plat_name.lower():
        if helics_version.startswith("3") and int(helics_version.split(".")[1]) >= 1:  # >= 3.1.x
            default_url = (
                "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-macOS-universal2.zip".format(
                    helics_version=helics_version
                )
            )
        else:
            default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-macOS-x86_64.zip".format(
                helics_version=helics_version
            )
    elif "win" in plat_name.lower():
        if "win32" in plat_name.lower():
            default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-win32.zip".format(
                helics_version=helics_version
            )
        else:
            default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-win64.zip".format(
                helics_version=helics_version
            )

    elif "linux" in plat_name.lower():
        default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-Linux-x86_64.tar.gz".format(
            helics_version=helics_version
        )
    elif platform.system() == "Darwin":
        if helics_version.startswith("3") and int(helics_version.split(".")[1]) >= 1:  # >= 3.1.x
            default_url = (
                "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-macOS-universal2.zip".format(
                    helics_version=helics_version
                )
            )
        else:
            default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-macOS-x86_64.zip".format(
                helics_version=helics_version
            )
    elif platform.system() == "Windows":
        if struct.calcsize("P") * 8 == 32:
            default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-win32.zip".format(
                helics_version=helics_version
            )
        else:
            default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-win64.zip".format(
                helics_version=helics_version
            )

    elif platform.system() == "Linux":
        default_url = "https://github.com/GMLC-TDC/HELICS/releases/download/v{helics_version}/Helics-{helics_version}-Linux-x86_64.tar.gz".format(
            helics_version=helics_version
        )
    else:
        raise NotImplementedError("Unsupported platform {}".format(platform.system()))

    return default_url


def _is_symlink(file_info):
    """
    Check the upper 4 bits of the external attribute for a symlink.
    See: https://unix.stackexchange.com/questions/14705/the-zip-formats-external-file-attribute
    Parameters
    ----------
    file_info : zipfile.ZipInfo
        The ZipInfo for a ZipFile
    Returns
    -------
    bool
        A response regarding whether the ZipInfo defines a symlink or not.
    """

    return (file_info.external_attr >> 28) == 0xA


def _extract(file_info, output_dir, zip_ref):
    """
    Unzip the given file into the given directory while preserving file permissions in the process.
    Parameters
    ----------
    file_info : zipfile.ZipInfo
        The ZipInfo for a ZipFile
    output_dir : str
        Path to the directory where the it should be unzipped to
    zip_ref : zipfile.ZipFile
        The ZipFile we are working with.
    Returns
    -------
    string
        Returns the target path the Zip Entry was extracted to.
    """

    # Handle any regular file/directory entries
    if not _is_symlink(file_info):
        return zip_ref.extract(file_info, output_dir)

    source = zip_ref.read(file_info.filename).decode("utf8")
    link_name = os.path.normpath(os.path.join(output_dir, file_info.filename))

    # make leading dirs if needed
    leading_dirs = os.path.dirname(link_name)
    if not os.path.exists(leading_dirs):
        os.makedirs(leading_dirs)

    # If the link already exists, delete it or symlink() fails
    if os.path.lexists(link_name):
        os.remove(link_name)

    # Create a symbolic link pointing to source named link_name.
    os.symlink(source, link_name)

    return link_name


def _set_permissions(zip_file_info, extracted_path):
    """
    Sets permissions on the extracted file by reading the ``external_attr`` property of given file info.
    Parameters
    ----------
    zip_file_info : zipfile.ZipInfo
        Object containing information about a file within a zip archive
    extracted_path : str
        Path where the file has been extracted to
    """

    # Permission information is stored in first two bytes.
    permission = zip_file_info.external_attr >> 16
    if not permission:
        return

    os.chmod(extracted_path, permission)


def _override_permissions(path, permission):
    """
    Forcefully override the permissions on the path
    Parameters
    ----------
    path str
        Path where the file or directory
    permission octal int
        Permission to set
    """
    if permission:
        os.chmod(path, permission)


def unzip(zip_file_path, output_dir, permission=None):
    """
    Unzip the given file into the given directory while preserving file permissions in the process.
    Parameters
    ----------
    zip_file_path : str
        Path to the zip file
    output_dir : str
        Path to the directory where the it should be unzipped to
    permission : int
        Permission to set in an octal int form
    """
    extracted_path = None
    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:

        # For each item in the zip file, extract the file and set permissions if available
        for file_info in zip_ref.infolist():
            extracted_path = _extract(file_info, output_dir, zip_ref)

            # If the extracted_path is a symlink, do not set the permissions. If the target of the symlink does not
            # exist, then os.chmod will fail with FileNotFoundError
            if not os.path.islink(extracted_path):
                _set_permissions(file_info, extracted_path)
                _override_permissions(extracted_path, permission)

    if extracted_path is not None and not os.path.islink(extracted_path):
        _override_permissions(output_dir, permission)


class HELICSDownloadCommand(Command):
    description = "Download helics libraries dependency"
    user_options = [
        ("pyhelics-install=", None, "path to pyhelics install folder"),
        ("plat-name=", None, "platform name to embed in generated filenames"),
    ]

    def initialize_options(self):
        self.plat_name = ""
        self.pyhelics_install = os.path.join(CURRENT_DIRECTORY, "./helics/install")
        if os.path.exists(self.pyhelics_install):
            shutil.rmtree(self.pyhelics_install)

    def finalize_options(self):
        pass

    def run(self):
        self.helics_url = create_default_url(HELICS_VERSION, self.plat_name)
        print("Downloading {}".format(self.helics_url))
        r = urlopen(self.helics_url)
        if r.getcode() == 200:
            if self.helics_url.endswith(".zip"):
                with open("./tmp.zip", "wb") as f:
                    f.write(r.read())
                unzip("./tmp.zip", self.pyhelics_install)
                os.remove("./tmp.zip")

                if len(os.listdir(self.pyhelics_install)) == 1 and os.listdir(self.pyhelics_install)[0].startswith("Helics-"):
                    tmp = os.listdir(self.pyhelics_install)[0]
                    for folder in os.listdir(os.path.join(self.pyhelics_install, tmp)):
                        p = Path(os.path.join(self.pyhelics_install, tmp, folder)).absolute()
                        parent_dir = p.parents[1]
                        p.rename(parent_dir / p.name)
            else:
                content = io.BytesIO(r.read())
                content.seek(0)
                with tarfile.open(fileobj=content) as tf:
                    dirname = tf.getnames()[0].partition("/")[0]
                    tf.extractall()
                shutil.move(dirname, self.pyhelics_install)
            for file in os.listdir(os.path.join(self.pyhelics_install, "bin")):
                f = Path(os.path.join(self.pyhelics_install, "bin", file))
                try:
                    import stat

                    f.chmod(f.stat().st_mode | stat.S_IEXEC)
                except:
                    pass
            files = [
                "helics_api.h",
                "helics_enums.h",
                os.path.join("shared_api_library", "api-data.h"),
                os.path.join("shared_api_library", "helics.h"),
                os.path.join("shared_api_library", "helics_export.h"),
                os.path.join("shared_api_library", "MessageFederate.h"),
                os.path.join("shared_api_library", "MessageFilters.h"),
                os.path.join("shared_api_library", "ValueFederate.h"),
                os.path.join("shared_api_library", "helicsCallbacks.h"),
            ]
            IGNOREBLOCK = False
            print("Writing to {}".format(os.path.abspath(self.pyhelics_install)))
            for file in files:
                if not os.path.isfile(os.path.join(self.pyhelics_install, "include", "helics", file)):
                    continue
                with open(os.path.join(self.pyhelics_install, "include", "helics", file)) as f:
                    lines = []
                    for line in f:
                        if line.startswith("#ifdef __cplusplus"):
                            IGNOREBLOCK = True
                            continue
                        if IGNOREBLOCK is True and line.startswith("#endif"):
                            IGNOREBLOCK = False
                            continue
                        if IGNOREBLOCK is True:
                            continue
                        if line.startswith("#"):
                            continue
                        lines.append(line)
                    data = "\n".join(lines)
                    data = data.replace("HELICS_EXPORT", "")
                    data = data.replace("HELICS_DEPRECATED_EXPORT", "")
                with open(os.path.join(self.pyhelics_install, "include", "helics", file), "w") as f:
                    f.write(data)


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=HELICS_SOURCE):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = sourcedir


class HELICSCMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
            cmake_version = re.search(r"version\s*([\d.]+)", out.decode().lower()).group(1)
            cmake_version = [int(i) for i in cmake_version.split(".")]
            if cmake_version < [3, 5, 1]:
                raise RuntimeError("CMake >= 3.5.1 is required to build helics")

        except OSError:
            if not os.path.exists(PYHELICS_INSTALL):
                raise RuntimeError("CMake must be installed to build the following extensions: " + ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        self.helics_url = DOWNLOAD_URL
        self.helics_source = HELICS_SOURCE
        if not os.path.exists(PYHELICS_INSTALL):
            r = urlopen(self.helics_url)
            if r.getcode() == 200:
                content = io.BytesIO(r.read())
                content.seek(0)
                with tarfile.open(fileobj=content) as tf:
                    tf.extractall(self.helics_source)
        else:
            return

        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        extdir = os.path.join(extdir, "helics", "install")
        # required for auto - detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cmake_args = [
            "-DHELICS_DISABLE_GIT_OPERATIONS=OFF",
            "-DCMAKE_BUILD_TYPE=Release",
            "-DCMAKE_INSTALL_PREFIX={}".format(extdir),
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ["-A", "x64"]
                build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j"]

        env = os.environ.copy()
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        cmd = " ".join(["cmake", "--build", ".", "--target", "install"] + build_args)
        print(cmd)
        subprocess.check_call(shlex.split(cmd), cwd=self.build_temp)

        files = [
            "helics_api.h",
            "helics_enums.h",
            os.path.join("shared_api_library", "api-data.h"),
            os.path.join("shared_api_library", "helics.h"),
            os.path.join("shared_api_library", "helics_export.h"),
            os.path.join("shared_api_library", "MessageFederate.h"),
            os.path.join("shared_api_library", "MessageFilters.h"),
            os.path.join("shared_api_library", "ValueFederate.h"),
            os.path.join("shared_api_library", "helicsCallbacks.h"),
        ]
        IGNOREBLOCK = False
        for file in files:
            if not os.path.isfile(os.path.join(extdir, "include", "helics", file)):
                continue
            with open(os.path.join(extdir, "include", "helics", file)) as f:
                lines = []
                for line in f:
                    if line.startswith("#ifdef __cplusplus"):
                        IGNOREBLOCK = True
                        continue
                    if IGNOREBLOCK is True and line.startswith("#endif"):
                        IGNOREBLOCK = False
                        continue
                    if IGNOREBLOCK is True:
                        continue
                    if line.startswith("#"):
                        continue
                    lines.append(line)
                data = "\n".join(lines)
                data = data.replace("HELICS_EXPORT", "")
                data = data.replace("HELICS_DEPRECATED_EXPORT", "")
            with open(os.path.join(extdir, "include", "helics", file), "w") as f:
                f.write(data)


install_requires = ["cffi>=1.6.0", "strip-hints", "click>=8"]

if sys.version_info < (3, 4):
    install_requires.append("enum34")


class HelicsBdistWheel(bdist_wheel):
    def get_tag(self):
        rv = bdist_wheel.get_tag(self)
        if platform.python_version().startswith("2"):
            return ("py2", "none") + rv[2:]
        else:
            return ("py3", "none") + rv[2:]


cmdclass = {
    "download": HELICSDownloadCommand,
    "build_ext": js_prerelease(HELICSCMakeBuild),
    "bdist_wheel": js_prerelease(HelicsBdistWheel),
    "develop": js_prerelease(develop),
    "build_py": js_prerelease(build_py),
    "egg_info": js_prerelease(egg_info),
    "sdist": js_prerelease(sdist, strict=True),
    "jsdeps": NPM,
}


class BinaryDistribution(Distribution):
    def is_pure(self):
        return False


helics_cli_install_requires = ["flask>=2", "requests", "flask-restful", "flask-cors", "pandas", "SQLAlchemy", "matplotlib"]

setup(
    name="helics",
    version=PYHELICS_VERSION,
    license="MIT",
    description="Python HELICS bindings",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Dheepak Krishnamurthy",
    author_email="me@kdheepak.com",
    url="https://github.com/GMLC-TDC/pyhelics",
    packages=["helics"],
    distclass=BinaryDistribution,
    py_modules=[splitext(basename(path))[0] for path in glob("helics/*.py")],
    # data_files=[("helics", ["install/include/helics/chelics.h"])],
    # cffi_modules=['helics/helics_build.py:ffibuilder'],
    package_data={"helics": ["install/*"]},
    ext_modules=[CMakeExtension("helics")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Utilities",
    ],
    project_urls={"Issue Tracker": "https://github.com/GMLC-TDC/pyhelics/issues"},
    keywords=["helics", "co-simulation"],
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*",
    install_requires=install_requires,
    extras_require={
        "cli": install_requires + helics_cli_install_requires,
        "tests": ["pytest", "pytest-ordering", "pytest-cov", "pytest-runner"],
        "docs": [
            "mkdocs",
            "inari[mkdocs]",
            "mkdocs-material",
            "black",
            "pygments",
            "pymdown-extensions",
        ],
    },
    cmdclass=cmdclass,
    entry_points={
        "console_scripts": [
            "helics=helics.cli:cli",
            "helics_app=helics.bin:helics_app",
            "helics_broker=helics.bin:helics_broker",
            "helics_broker_server=helics.bin:helics_broker_server",
            "helics_player=helics.bin:helics_player",
            "helics_recorder=helics.bin:helics_recorder",
        ]
    },
)



# -*- coding: utf-8 -*-
import os
import sys

CURRENT_DIRECTORY = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(CURRENT_DIRECTORY)
sys.path.append(os.path.dirname(CURRENT_DIRECTORY))

import time
import helics as h

from test_init import createBroker, createValueFederate, destroyFederate, destroyBroker, createMessageFederate


def test_broker():
    broker = createBroker(1)
    initstring = "--broker="
    identifier = h.helicsBrokerGetIdentifier(broker)
    initstring = initstring + identifier
    initstring = initstring + " --broker_address "
    address = h.helicsBrokerGetAddress(broker)
    initstring = initstring + address
    assert initstring == "--broker=mainbroker --broker_address tcp://127.0.0.1:23404"
    destroyBroker(broker)


def test_messagefilter_registration():

    broker = createBroker(2)

    fFed, ffedinfo = createMessageFederate(1, "filter")
    mFed, mfedinfo = createMessageFederate(1, "message")

    h.helicsFederateRegisterGlobalEndpoint(mFed, "port1", "")
    h.helicsFederateRegisterGlobalEndpoint(mFed, "port2", "")

    f1 = h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "filter1")
    f2 = h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "filter2")
    h.helicsFederateRegisterEndpoint(fFed, "fout", "")
    h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "filter0/fout")
    h.helicsFederateEnterExecutingModeAsync(fFed)
    h.helicsFederateEnterExecutingMode(mFed)
    h.helicsFederateEnterExecutingModeComplete(fFed)

    filter_name = h.helicsFilterGetName(f1)
    assert filter_name == "Testfilter/filter1"

    filter_name = h.helicsFilterGetName(f2)
    assert filter_name == "Testfilter/filter2"

    # filter_target = h.helicsFilterGetTarget(f2)
    # assert filter_target == "port2"

    h.helicsFederateDisconnect(mFed)
    h.helicsFederateDisconnect(fFed)

    destroyFederate(fFed, ffedinfo)
    destroyFederate(mFed, mfedinfo)
    time.sleep(1.0)

    destroyBroker(broker)


def test_messagefilter_info():

    broker = createBroker(2)

    fFed, ffedinfo = createMessageFederate(1, "filter")
    mFed, mfedinfo = createMessageFederate(1, "message")

    p1 = h.helicsFederateRegisterGlobalEndpoint(mFed, "port1", "")
    p2 = h.helicsFederateRegisterGlobalEndpoint(mFed, "port2", "")

    h.helicsEndpointSetInfo(p1, "p1_test")
    h.helicsEndpointSetInfo(p2, "p2_test")

    f1 = h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "filter1")
    h.helicsFilterAddSourceTarget(f1, "port1")
    h.helicsFilterSetInfo(f1, "f1_test")

    f2 = h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "filter2")
    h.helicsFilterAddSourceTarget(f2, "port2")
    h.helicsFilterSetInfo(f2, "f2_test")

    ep1 = h.helicsFederateRegisterEndpoint(fFed, "fout", "")
    h.helicsEndpointSetInfo(ep1, "ep1_test")

    f3 = h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "c4")
    h.helicsFilterAddSourceTarget(f3, "filter0/fout")
    h.helicsFilterSetInfo(f3, "f3_test")

    h.helicsFederateEnterExecutingModeAsync(fFed)
    h.helicsFederateEnterExecutingMode(mFed)
    h.helicsFederateEnterExecutingModeComplete(fFed)

    filter_name = h.helicsFilterGetName(f1)
    assert filter_name == "Testfilter/filter1"

    filter_name = h.helicsFilterGetName(f2)
    assert filter_name == "Testfilter/filter2"

    assert h.helicsEndpointGetInfo(p1) == "p1_test"
    assert h.helicsEndpointGetInfo(p2) == "p2_test"
    assert h.helicsEndpointGetInfo(ep1) == "ep1_test"

    assert h.helicsFilterGetInfo(f1) == "f1_test"
    assert h.helicsFilterGetInfo(f2) == "f2_test"
    assert h.helicsFilterGetInfo(f3) == "f3_test"

    h.helicsFederateDisconnect(mFed)
    h.helicsFederateDisconnect(fFed)

    destroyFederate(fFed, ffedinfo)
    destroyFederate(mFed, mfedinfo)
    time.sleep(1.0)

    destroyBroker(broker)


def test_messagefilter_function():
    broker = createBroker(2)

    fFed, ffedinfo = createMessageFederate(1, "filter")
    mFed, mfedinfo = createMessageFederate(1, "message")

    p1 = h.helicsFederateRegisterGlobalEndpoint(mFed, "port1", "")
    p2 = h.helicsFederateRegisterGlobalEndpoint(mFed, "port2", "random")

    f1 = h.helicsFederateRegisterGlobalFilter(fFed, h.HELICS_FILTER_TYPE_CUSTOM, "filter1")
    h.helicsFilterAddSourceTarget(f1, "port1")
    f2 = h.helicsFederateRegisterGlobalFilter(fFed, h.HELICS_FILTER_TYPE_DELAY, "filter2")
    h.helicsFilterAddSourceTarget(f2, "port1")
    h.helicsFederateRegisterEndpoint(fFed, "fout", "")
    f3 = h.helicsFederateRegisterFilter(fFed, h.HELICS_FILTER_TYPE_RANDOM_DELAY, "filter3")
    h.helicsFilterAddSourceTarget(f3, "filter/fout")

    h.helicsFilterSet(f2, "delay", 2.5)
    h.helicsFederateEnterExecutingModeAsync(fFed)
    h.helicsFederateEnterExecutingMode(mFed)
    h.helicsFederateEnterExecutingModeComplete(fFed)
    state = h.helicsFederateGetState(fFed)
    assert state == 2
    data = "hello world"

    filt_key = h.helicsFilterGetName(f1)
    assert filt_key == "filter1"

    filt_key = h.helicsFilterGetName(f2)
    assert filt_key == "filter2"

    h.helicsEndpointSendBytesTo(p1, data.encode(), "port2")
    h.helicsFederateRequestTimeAsync(mFed, 1.0)
    grantedtime = h.helicsFederateRequestTime(fFed, 1.0)
    assert grantedtime == 1.0
    grantedtime = h.helicsFederateRequestTimeComplete(mFed)
    assert grantedtime == 1.0
    res = h.helicsFederateHasMessage(mFed)
    assert res == 0
    res = h.helicsEndpointHasMessage(p2)
    assert res == 0
    # grantedtime = h.helicsFederateRequestTime(fFed, 3.0)
    # assert res==h.helics_true

    h.helicsFederateDisconnect(mFed)
    h.helicsFederateDisconnect(fFed)
    # f2 = h.helicsFederateRegisterDestinationFilter(fFed, h.helics_custom_filter, "filter2", "port2")
    # ep1 = h.helicsFederateRegisterEndpoint(fFed, "fout", "")
    # f3 = h.helicsFederateRegisterSourceFilter(fFed, h.helics_custom_filter, "", "filter0/fout")

    destroyFederate(fFed, ffedinfo)
    destroyFederate(mFed, mfedinfo)
    time.sleep(1.0)
    destroyBroker(broker)
